"""
Status factory module.
"""
from __future__ import annotations

import typing

from digitalhub_core.entities._base.status import State
from digitalhub_core.entities.artifacts.status import ArtifactStatus
from digitalhub_core.entities.dataitems.status import DataitemStatus
from digitalhub_core.entities.functions.status import FunctionStatus
from digitalhub_core.entities.projects.status import ProjectStatus
from digitalhub_core.entities.runs.status import RunStatus
from digitalhub_core.entities.tasks.status import TaskStatus
from digitalhub_core.entities.workflows.status import WorkflowStatus
from digitalhub_core.utils.commons import ARTF, DTIT, FUNC, PROJ, RUNS, TASK, WKFL

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.status import Status


class StatusBuilder(dict):
    """
    Status builder class.
    """

    def register(self, module: str, status: Status) -> None:
        """
        Register status.

        Parameters
        ----------
        module: str
            module name.
        status: Status
            Status object.

        Returns
        -------
        None
        """
        self[module] = status

    def build(self, module: str, **kwargs) -> Status:
        """
        Build entity status object.

        Parameters
        ----------
        module: str
            module name.
        **kwargs
            Keyword arguments.

        Returns
        -------
        Status
            An entity status object.
        """
        if module not in self:
            raise ValueError(f"Invalid module name: {module}")
        kwargs = self._parse_arguments(**kwargs)
        return self[module](**kwargs)

    @staticmethod
    def _parse_arguments(**kwargs) -> dict:
        """
        Parse keyword arguments and add default values.

        Parameters
        ----------
        **kwargs
            Keyword arguments.

        Returns
        -------
        dict
            Keyword arguments with default values.
        """
        state = kwargs.get("state")
        if state is None:
            kwargs["state"] = State.CREATED.value
        else:
            if kwargs["state"] not in State.__members__:
                raise ValueError(f"Invalid state: {state}")
        return kwargs


def build_status(module: str, **kwargs) -> Status:
    """
    Wrapper for StatusBuilder.build.

    Parameters
    ----------
    module: str
        module name.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Status
        An entity status object.
    """
    return status_builder.build(module, **kwargs)


status_builder = StatusBuilder()
status_builder.register(ARTF, ArtifactStatus)
status_builder.register(DTIT, DataitemStatus)
status_builder.register(FUNC, FunctionStatus)
status_builder.register(PROJ, ProjectStatus)
status_builder.register(RUNS, RunStatus)
status_builder.register(TASK, TaskStatus)
status_builder.register(WKFL, WorkflowStatus)



"""
Runtime factory module.
"""
from __future__ import annotations

import importlib
import typing

from digitalhub_core.utils.modules_utils import import_registry

if typing.TYPE_CHECKING:
    from digitalhub_core.runtimes.base import Runtime


def build_runtime(kind: str) -> Runtime:
    """
    Build runtimes.

    Parameters
    ----------
    kind : str
        The runtime kind.

    Returns
    -------
    Runtime
        Runtime instance.
    """
    try:
        registry = import_registry(kind)
        module = importlib.import_module(registry.get_runtime().module)
        return getattr(module, registry.get_runtime().class_name)()
    except (ModuleNotFoundError, ImportError):
        raise ValueError(f"Runtime {kind} not found")



"""
Store module.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from tempfile import mkdtemp
from typing import Literal

import pandas as pd
from digitalhub_core.utils.exceptions import StoreError
from digitalhub_core.utils.file_utils import build_path, make_dir
from digitalhub_core.utils.uri_utils import get_name_from_uri, map_uri_scheme
from pydantic import BaseModel


class Store(metaclass=ABCMeta):
    """
    Store abstract class.
    """

    def __init__(self, name: str, store_type: str) -> None:
        """
        Constructor.

        Parameters
        ----------
        name : str
            Store name.
        store_type : str
            Store type. Used to choose the right store implementation.
        uri : str
            Store URI.

        Returns
        -------
        None
        """
        self.name = name
        self.type = store_type

        # Private attributes
        self._registry: dict[str, str] = {}

    ############################
    # IO methods
    ############################

    @abstractmethod
    def download(self, src: str, dst: str | None = None) -> str:
        """
        Method to download artifact from storage.
        """

    @abstractmethod
    def fetch_artifact(self, src: str, dst: str | None = None) -> str:
        """
        Method to fetch artifact from storage.
        """

    @abstractmethod
    def upload(self, src: str, dst: str | None = None) -> str:
        """
        Method to upload artifact to storage.
        """

    @abstractmethod
    def persist_artifact(self, src: str, dst: str | None = None) -> str:
        """
        Method to persist artifact in storage.
        """

    @abstractmethod
    def write_df(self, df: pd.DataFrame, dst: str | None = None, **kwargs) -> str:
        """
        Write pandas DataFrame as parquet or csv.
        """

    @staticmethod
    def read_df(path: str, extension: str, **kwargs) -> pd.DataFrame:
        """
        Read DataFrame from path.

        Parameters
        ----------
        path : str
            Path to read DataFrame from.
        extension : str
            Extension of the file.
        **kwargs
            Keyword arguments.

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame.

        Raises
        ------
        ValueError
            If format is not supported.
        """
        if extension == "csv":
            return pd.read_csv(path, **kwargs)
        if extension == "parquet":
            return pd.read_parquet(path, **kwargs)
        raise ValueError(f"Format {extension} not supported.")

    ############################
    # Helpers methods
    ############################

    @staticmethod
    def _check_local_dst(dst: str) -> None:
        """
        Check if the local destination directory exists. Create in case it does not.

        Parameters
        ----------
        dst : str
            The destination directory.

        Returns
        -------
        None

        Raises
        ------
        StoreError
            If the destination is not a local path.
        """
        if map_uri_scheme(dst) != "local":
            raise StoreError(f"Destination '{dst}' is not a local path.")
        make_dir(dst)

    def _build_temp(self, src: str) -> str:
        """
        Build a temporary path.

        Parameters
        ----------
        src : str
            Source filename.

        Returns
        -------
        str
            Temporary path.
        """
        tmpdir = mkdtemp()
        dst = build_path(tmpdir, get_name_from_uri(src))
        self._registry[src] = dst
        return dst

    @staticmethod
    @abstractmethod
    def is_local() -> bool:
        """
        Method to check if store is local.
        """


class StoreConfig(BaseModel):
    """
    Store configuration base class.
    """


class StoreParameters(BaseModel):
    """
    Store configuration class.
    """

    name: str
    """Store id."""

    type: Literal["local", "s3", "remote", "sql"]
    """Store type to instantiate."""

    config: StoreConfig = None
    """Configuration for the store."""

    is_default: bool = False
    """Flag to determine if the store is the default one."""



"""
Exceptions module.
"""
from __future__ import annotations


class GenericError(Exception):
    """
    Base class for exception.
    """


class StoreError(GenericError):
    """
    Raised when incontered errors on Stores.
    """


class BackendError(GenericError):
    """
    Raised when incontered errors from backend.
    """


class EntityError(GenericError):
    """
    Raised when incontered errors on Entities.
    """



"""
Project operations module.
"""
from __future__ import annotations

import typing

from digitalhub_core.client.builder import get_client
from digitalhub_core.context.builder import delete_context
from digitalhub_core.entities.projects.entity import project_from_dict, project_from_parameters
from digitalhub_core.utils.api import api_base_delete, api_base_read, api_base_update
from digitalhub_core.utils.commons import PROJ
from digitalhub_core.utils.exceptions import BackendError, EntityError
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.projects.entity import Project


def create_project(**kwargs) -> Project:
    """
    Create a new project.

    Parameters
    ----------
    **kwargs
        Keyword arguments.

    Returns
    -------
    Project
        A Project instance.
    """
    return project_from_parameters(**kwargs)


def new_project(
    name: str,
    uuid: str | None = None,
    description: str | None = None,
    local: bool = False,
    context: str = "",
    source: str = "",
    **kwargs,
) -> Project:
    """
    Create project.

    Parameters
    ----------
    name : str
        Identifier of the project.
    kind : str
        The type of the project.
    uuid : str
        UUID.
    description : str
        Description of the project.
    local : bool
        Flag to determine if object will be exported to backend.
    context : str
        The context of the project.
    source : str
        The source of the project.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Project
        Project object.
    """
    obj = create_project(
        name=name,
        kind="project",
        uuid=uuid,
        description=description,
        local=local,
        context=context,
        source=source,
        **kwargs,
    )
    obj.save()
    return obj


def load_project(name: str | None = None, filename: str | None = None, local: bool = False) -> Project:
    """
    Load project and context from backend or file.

    Parameters
    ----------
    name : str
        Name of the project.
    filename : str
        Path to file where to load project from.
    local : bool
        Flag to determine if backend is local.

    Returns
    -------
    Project
        A Project instance with setted context.
    """
    if name is not None:
        return get_project(name, local)
    if filename is not None:
        return import_project(filename, local)
    raise EntityError("Either name or filename must be provided.")


def get_or_create_project(
    name: str,
    local: bool = False,
    **kwargs,
) -> Project:
    """
    Get or create a project.

    Parameters
    ----------
    name : str
        Name of the project.
    local : bool
        Flag to determine if backend is local.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Project
        A Project instance.
    """
    try:
        return get_project(name, local)
    except BackendError:
        return new_project(name, local=local, **kwargs)


def get_project(name: str, local: bool = False) -> Project:
    """
    Retrieves project details from the backend.

    Parameters
    ----------
    name : str
        The name or UUID.
    local : bool
        Flag to determine if backend is local.

    Returns
    -------
    Project
        Object instance.
    """
    api = api_base_read(PROJ, name)
    client = get_client(local)
    obj = client.read_object(api)
    return project_from_dict(obj)


def import_project(file: str, local: bool = False) -> Project:
    """
    Import an Project object from a file using the specified file path.

    Parameters
    ----------
    file : str
        Path to the file.
    local : bool
        Flag to determine if backend is local.

    Returns
    -------
    Project
        Object instance.
    """
    obj = read_yaml(file)
    obj["local"] = local
    return project_from_dict(obj)


def delete_project(name: str, local: bool = False) -> list[dict]:
    """
    Delete a project.

    Parameters
    ----------
    name : str
        Name of the project.
    local : bool
        Flag to determine if backend is local.

    Returns
    -------
    dict
        Response from backend.
    """
    client = get_client(local)
    api = api_base_delete(PROJ, name)
    response = client.delete_object(api)
    delete_context(name)
    return response


def update_project(project: Project, local: bool = False) -> dict:
    """
    Update a project.

    Parameters
    ----------
    project : Project
        The project to update.
    local : bool
        Flag to determine if backend is local.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_base_update(PROJ, project.metadata.name)
    client = get_client(local)
    return client.update_object(project.to_dict(), api)



"""
Module for performing operations on tasks.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities.tasks.entity import task_from_dict, task_from_parameters
from digitalhub_core.utils.api import api_base_delete, api_base_read, api_base_update
from digitalhub_core.utils.commons import TASK
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.tasks.entity import Task


def create_task(**kwargs) -> Task:
    """
    Create a new object instance.

    Parameters
    ----------
    **kwargs
        Keyword arguments.

    Returns
    -------
    Task
       Object instance.
    """
    return task_from_parameters(**kwargs)


def new_task(
    project: str,
    kind: str,
    uuid: str | None = None,
    function: str | None = "",
    **kwargs,
) -> Task:
    """
    Create a new object instance.

    Parameters
    ----------
    project : str
        Name of the project.
    kind : str
        The type of the task.
    uuid : str
        UUID.
    function : str
        The function string identifying the function.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Task
       Object instance.
    """
    obj = create_task(
        project=project,
        kind=kind,
        uuid=uuid,
        function=function,
        **kwargs,
    )
    obj.save()
    return obj


def get_task(project: str, name: str) -> Task:
    """
    Get object from backend.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the task.

    Returns
    -------
    Task
        Object instance.
    """
    api = api_base_read(TASK, name)
    obj = get_context(project).read_object(api)
    return task_from_dict(obj)


def import_task(file: str) -> Task:
    """
    Get object from file.

    Parameters
    ----------
    file : str
        Path to the file.

    Returns
    -------
    Task
        Object instance.
    """
    obj = read_yaml(file)
    return task_from_dict(obj)


def delete_task(project: str, name: str) -> dict:
    """
    Delete task from the backend.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the task.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_base_delete(TASK, name)
    return get_context(project).delete_object(api)


def update_task(task: Task) -> dict:
    """
    Update task.

    Parameters
    ----------
    task : Task
        The task object to update.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_base_update(TASK, task.id)
    return get_context(task.metadata.project).update_object(task.to_dict(), api)



"""
WorkflowStatus class module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.status import Status


class WorkflowStatus(Status):
    """
    Status class for workflow entities.
    """



"""
Module for performing operations on runs.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities.runs.entity import run_from_dict, run_from_parameters
from digitalhub_core.utils.api import api_base_delete, api_base_read, api_base_update
from digitalhub_core.utils.commons import RUNS
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.runs.entity import Run


def create_run(**kwargs) -> Run:
    """
    Create a new object instance.

    Parameters
    ----------
    **kwargs
        Keyword arguments.

    Returns
    -------
    Run
       Object instance.
    """
    return run_from_parameters(**kwargs)


def new_run(
    project: str,
    task: str,
    task_id: str,
    kind: str,
    uuid: str | None = None,
    inputs: dict | None = None,
    outputs: list | None = None,
    parameters: dict | None = None,
    local_execution: bool = False,
    **kwargs,
) -> Run:
    """
    Create run.

    Parameters
    ----------
    project : str
        Name of the project.
    task_id : str
        Identifier of the task associated with the run.
    task : str
        Name of the task associated with the run.
    kind : str
        The type of the run.
    uuid : str
        UUID.
    inputs : dict
        Inputs of the run.
    outputs : list
        Outputs of the run.
    parameters : dict
        Parameters of the run.
    local_execution : bool
        Flag to determine if object has local execution.
    embedded : bool
        Flag to determine if object must be embedded in project.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Run
        Run object.
    """
    obj = create_run(
        project=project,
        task=task,
        task_id=task_id,
        kind=kind,
        uuid=uuid,
        inputs=inputs,
        outputs=outputs,
        parameters=parameters,
        local_execution=local_execution,
        **kwargs,
    )
    obj.save()
    return obj


def get_run(project: str, name: str) -> Run:
    """
    Get object from backend.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the run.

    Returns
    -------
    Run
        Object instance.
    """
    api = api_base_read(RUNS, name)
    obj = get_context(project).read_object(api)
    return run_from_dict(obj)


def import_run(file: str) -> Run:
    """
    Get object from file.

    Parameters
    ----------
    file : str
        Path to the file.

    Returns
    -------
    Run
        Object instance.
    """
    obj = read_yaml(file)
    return run_from_dict(obj)


def delete_run(project: str, name: str) -> dict:
    """
    Delete run from the backend.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the run.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_base_delete(RUNS, name)
    return get_context(project).delete_object(api)


def update_run(run: Run) -> dict:
    """
    Update run.

    Parameters
    ----------
    run : Run
        The run object to update.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_base_update(RUNS, run.id)
    return get_context(run.metadata.project).update_object(run.to_dict(include_all_non_private=True), api)



"""
Common IO utils.
"""
from __future__ import annotations

import json
import shutil
from collections import OrderedDict
from io import BufferedReader, BytesIO, StringIO, TextIOBase, TextIOWrapper
from pathlib import Path
from typing import IO

import yaml

####################
# Wrappers
####################


#  https://stackoverflow.com/questions/55889474/convert-io-stringio-to-io-bytesio
#  made by foobarna, improved by imporsen
#  Documentation added with copilot
class BytesIOWrapper(BufferedReader):
    """
    Wrap a buffered bytes stream over TextIOBase string stream.
    """

    def __init__(
        self,
        text_io_buffer: TextIOBase,
        encoding: str | None = None,
        errors: str | None = None,
        **kwargs,
    ) -> None:
        """
        Keyword arguments.

        Wrap a buffered bytes stream over TextIOBase string stream.

        Parameters
        ----------
        text_io_buffer : TextIOBase
            The string stream to wrap.
        encoding : str
            The encoding to use, by default None
        errors : str
            ...
        **kwargs
            Keyword arguments.
        """
        super().__init__(text_io_buffer, **kwargs)
        self.encoding = encoding or text_io_buffer.encoding or "utf-8"
        self.errors = errors or text_io_buffer.errors or "strict"

    def _encoding_call(self, method_name: str, *args, **kwargs) -> bytes:
        """
        Call a method of the raw stream, and encode the result.

        Parameters
        ----------
        method_name : str
            The name of the method to call.
        *args : list
            Arguments list.
        **kwargs
            Keyword arguments.

        Returns
        -------
        bytes
            The encoded result of the method call.
        """
        raw_method = getattr(self.raw, method_name)
        val = raw_method(*args, **kwargs)
        return val.encode(self.encoding, errors=self.errors)

    def read(self, size: int | None = -1) -> bytes:
        """
        Read bytes from the stream.

        Parameters
        ----------
        size : int
            The number of bytes to read.

        Returns
        -------
        bytes
            The read bytes.
        """
        return self._encoding_call("read", size)

    def read1(self, size: int = -1) -> bytes:
        """
        Read bytes from the stream.

        Parameters
        ----------
        size : int
            The number of bytes to read.

        Returns
        -------
        bytes
            The read bytes.
        """
        return self._encoding_call("read1", size)

    def peek(self, size: int = 0) -> bytes:
        """
        Peek bytes from the stream.

        Parameters
        ----------
        size : int
            The number of bytes to peek.

        Returns
        -------
        bytes
            The peeked bytes.
        """
        return self._encoding_call("peek", size)


def wrap_bytes(src: BytesIO | StringIO) -> TextIOWrapper | StringIO:
    """
    Wrap a BytesIO in a StringIO.

    Parameters
    ----------
    src : BytesIO | StringIO
        The source object to be wrapped.

    Returns
    -------
    TextIOWrapper | StringIO
        The wrapped object.
    """
    if isinstance(src, BytesIO):
        return TextIOWrapper(src)
    return src


def wrap_string(src: BytesIO | StringIO) -> BytesIOWrapper | BytesIO:
    """
    Wrap a StringIO in a BytesIO.

    Parameters
    ----------
    src : BytesIO | StringIO
        The source object to be wrapped.

    Returns
    -------
    BytesIOWrapper | BytesIO
        The wrapped object.
    """
    if isinstance(src, StringIO):
        return BytesIOWrapper(src)
    return src


####################
# Writers
####################


def write_stringio(src: str) -> StringIO:
    """
    Write string in TextStream StringIO.

    Parameters
    ----------
    src : str
        The source string to be wrapped.

    Returns
    -------
    StringIO
        The wrapped object.
    """
    stringio = StringIO()
    stringio.write(src)
    stringio.seek(0)
    return stringio


def write_bytesio(src: str) -> BytesIO:
    """
    Write string in ByteStream BytesIO.

    Parameters
    ----------
    src : str
        The source string to be wrapped.

    Returns
    -------
    BytesIO
        The wrapped object.
    """
    bytesio = BytesIO()
    bytesio.write(src.encode())
    bytesio.seek(0)
    return bytesio


def write_json(data: dict, path: str | Path) -> None:
    """
    Store JSON file.

    Parameters
    ----------
    data : dict
        The data to be stored.
    path : str | Path
        The path to the file.

    Returns
    -------
    None
    """
    with open(path, "w", encoding="utf-8") as file:
        json.dump(data, file)


def write_text(string: str, path: str | Path) -> None:
    """
    Write text on a file.

    Parameters
    ----------
    string : str
        The string to be written.
    path : str | Path
        The path to the file.

    Returns
    -------
    None
    """
    with open(path, "w", encoding="utf-8") as file:
        file.write(string)


def write_bytes(byt: bytes, path: str | Path) -> None:
    """
    Write bytes on a file.

    Parameters
    ----------
    byt : bytes
        The bytes to be written.
    path : str | Path
        The path to the file.

    Returns
    -------
    None
    """
    with open(path, "wb") as file:
        file.write(byt)


def write_object(buff: IO, dst: str) -> None:
    """
    Write a buffer as file.

    Parameters
    ----------
    buff : IO
        The buffer to be written.
    dst : str
        The path to the file.

    Returns
    -------
    None
    """
    buff.seek(0)
    write_mode = "wb" if isinstance(buff, BytesIO) else "w"
    with open(dst, write_mode) as file:
        shutil.copyfileobj(buff, file)


def write_yaml(obj: dict, file: str | Path) -> None:
    """
    Write a dict to a yaml file.

    Parameters
    ----------
    obj : dict
        The dict to write.
    file : str | Path
        The yaml file path to write.

    Returns
    -------
    None
    """

    # Function needed to preserve the order of the keys in the yaml file.
    def ordered_dict_representer(self, value: OrderedDict) -> dict:
        """
        Represent an OrderedDict as a dict.

        Parameters
        ----------
        self : yaml.representer
            The yaml representer.
        value : OrderedDict
            The OrderedDict to represent.

        Returns
        -------
        dict
            The OrderedDict as a dict.
        """
        return self.represent_mapping("tag:yaml.org,2002:map", value.items())

    yaml.add_representer(OrderedDict, ordered_dict_representer)

    obj = OrderedDict(obj)
    with open(file, "w", encoding="utf-8") as out_file:
        yaml.dump(obj, out_file)


####################
# Readers
####################


def read_yaml(file: str | Path) -> dict:
    """
    Read a yaml file and return a dict.

    Parameters
    ----------
    file : str | Path
        The yaml file path to read.

    Returns
    -------
    dict
        The yaml file content.
    """
    with open(file, "r", encoding="utf-8") as in_file:
        data = yaml.load(in_file, Loader=yaml.SafeLoader)
    return data


def read_bytes(file: str | Path) -> bytes:
    """
    Read a file and return the bytes.

    Parameters
    ----------
    file : str | Path
        The file path to read.

    Returns
    -------
    bytes
        The file content.
    """
    with open(file, "rb") as in_file:
        data = in_file.read()
    return data


def read_text(file: str | Path) -> str:
    """
    Read a file and return the text.

    Parameters
    ----------
    file : str | Path
        The file path to read.

    Returns
    -------
    str
        The file content.
    """
    with open(file, "r", encoding="utf-8") as in_file:
        data = in_file.read()
    return data



"""
Entity metadata module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.base import ModelObj


class Metadata(ModelObj):
    """
    A class representing the metadata of an entity.
    """

    def __init__(self, project: str, created: str, updated: str) -> None:
        """
        Constructor.

        Parameters
        ----------
        project : str
            Name of the project.
        created : str
            Created date.
        updated : str
            Updated date.
        """
        self.project = project
        self.created = created
        self.updated = updated

    @classmethod
    def from_dict(cls, obj: dict) -> "Metadata":
        """
        Return entity metadata object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity metadata.

        Returns
        -------
        Metadata
            An entity metadata object.
        """
        return cls(**obj)



"""
Function module.
"""
from __future__ import annotations

import typing
from concurrent.futures import ThreadPoolExecutor

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities._base.entity import Entity
from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.entities.tasks.crud import create_task, delete_task, new_task
from digitalhub_core.utils.api import api_ctx_create, api_ctx_update
from digitalhub_core.utils.commons import FUNC
from digitalhub_core.utils.exceptions import EntityError
from digitalhub_core.utils.generic_utils import build_uuid, get_timestamp
from digitalhub_core.utils.io_utils import write_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.context.context import Context
    from digitalhub_core.entities.functions.metadata import FunctionMetadata
    from digitalhub_core.entities.functions.spec import FunctionSpec
    from digitalhub_core.entities.functions.status import FunctionStatus
    from digitalhub_core.entities.runs.entity import Run
    from digitalhub_core.entities.tasks.entity import Task


class Function(Entity):
    """
    A class representing a function.
    """

    def __init__(
        self,
        uuid: str,
        kind: str,
        metadata: FunctionMetadata,
        spec: FunctionSpec,
        status: FunctionStatus,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        uuid : str
            UUID.
        kind : str
            Kind of the object.
        metadata : FunctionMetadata
            Metadata of the object.
        spec : FunctionSpec
            Specification of the object.
        status : FunctionStatus
            Status of the object.
        """
        super().__init__()
        self.id = uuid
        self.kind = kind
        self.metadata = metadata
        self.spec = spec
        self.status = status

        # Private attributes
        self._tasks: dict[str, Task] = {}

        self.project = self.metadata.project
        self.name = self.metadata.name
        self.embedded = self.metadata.embedded
        self._obj_attr.extend(["project", "name", "embedded"])

    #############################
    #  Save / Export
    #############################

    def save(self, update: bool = False) -> dict:
        """
        Save function into backend.

        Parameters
        ----------
        uuid : str
            Specify uuid for the function update

        Returns
        -------
        dict
            Mapping representation of Function from backend.
        """
        obj = self.to_dict(include_all_non_private=True)

        if not update:
            api = api_ctx_create(self.metadata.project, FUNC)
            return self._context().create_object(obj, api)

        self.metadata.updated = obj["metadata"]["updated"] = get_timestamp()
        api = api_ctx_update(self.metadata.project, FUNC, self.metadata.name, self.id)
        return self._context().update_object(obj, api)

    def export(self, filename: str | None = None) -> None:
        """
        Export object as a YAML file.

        Parameters
        ----------
        filename : str
            Name of the export YAML file. If not specified, the default value is used.

        Returns
        -------
        None
        """
        obj = self.to_dict()
        filename = filename if filename is not None else f"function_{self.metadata.project}_{self.metadata.name}.yaml"
        write_yaml(filename, obj)

    #############################
    #  Context
    #############################

    def _context(self) -> Context:
        """
        Get context.

        Returns
        -------
        Context
            Context.
        """
        return get_context(self.metadata.project)

    #############################
    #  Function Methods
    #############################

    def run(
        self,
        action: str,
        inputs: dict | None = None,
        outputs: dict | None = None,
        parameters: dict | None = None,
        local_execution: bool = False,
        **kwargs,
    ) -> Run:
        """
        Run function.

        Parameters
        ----------
        action : str
            Action to execute. Task parameter.
        inputs : dict
            Function inputs. Run parameter.
        outputs : dict
            Function outputs. Run parameter.
        parameters : dict
            Function parameters. Run parameter.
        local_execution : bool
            Flag to determine if object has local execution. Run parameter.
        **kwargs
            Keyword arguments passed to Task builder.
        Returns
        -------
        Run
            Run instance.
        """

        # Create task if does not exists
        task = self._tasks.get(action)
        if task is None:
            task = self.new_task(kind=action, **kwargs)

        # Run function from task
        run = task.run(inputs, outputs, parameters, local_execution)

        # If execution is done by backend, return run
        if not local_execution:
            return run

        # If local execution, build run and run it
        run.build()
        with ThreadPoolExecutor(max_workers=1) as executor:
            result = executor.submit(run.run)
        return result.result()

    def _get_function_string(self) -> str:
        """
        Get function string.

        Returns
        -------
        str
            Function string.
        """
        return f"{self.kind}://{self.metadata.project}/{self.metadata.name}:{self.id}"

    #############################
    #  CRUD Methods for Tasks
    #############################

    def new_task(self, **kwargs) -> Task:
        """
        Create new task.

        Parameters
        ----------
        **kwargs
            Keyword arguments.

        Returns
        -------
        Task
            New task.
        """
        kwargs["project"] = self.metadata.project
        kwargs["function"] = self._get_function_string()
        task = new_task(**kwargs)
        self._tasks[kwargs["kind"]] = task
        return task

    def update_task(self, kind: str, **kwargs) -> None:
        """
        Update task.

        Parameters
        ----------
        kind : str
            Kind of the task.
        **kwargs
            Keyword arguments.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If task does not exist.
        """
        self._raise_if_not_exists(kind)

        # Update kwargs
        kwargs["project"] = self.metadata.project
        kwargs["kind"] = kind
        kwargs["function"] = self._get_function_string()
        kwargs["uuid"] = self._tasks[kind].id

        # Update task
        task = create_task(**kwargs)
        task.save(update=True)
        self._tasks[kind] = task

    def get_task(self, kind: str) -> Task:
        """
        Get task.

        Parameters
        ----------
        kind : str
            Kind of the task.

        Returns
        -------
        Task
            Task.

        Raises
        ------
        EntityError
            If task is not created.
        """
        self._raise_if_not_exists(kind)
        return self._tasks[kind]

    def delete_task(self, kind: str) -> None:
        """
        Delete task.

        Parameters
        ----------
        kind : str
            Kind of the task.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If task is not created.
        """
        self._raise_if_not_exists(kind)
        delete_task(self.metadata.project, self._tasks[kind].name)
        self._tasks[kind] = None

    def _raise_if_not_exists(self, kind: str) -> None:
        """
        Raise error if task is not created.

        Parameters
        ----------
        kind : str
            Kind of the task.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If task does not exist.
        """
        if self._tasks.get(kind) is None:
            raise EntityError("Task does not exist.")

    #############################
    #  Overridden Methods
    #############################

    @staticmethod
    def _parse_dict(entity: str, obj: dict) -> dict:
        """
        Get dictionary and parse it to a valid entity dictionary.

        Parameters
        ----------
        entity : str
            Entity type.
        obj : dict
            Dictionary to parse.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        uuid = build_uuid(obj.get("id"))
        kind = obj.get("kind", "")
        metadata = build_metadata(entity, **obj.get("metadata"))
        spec = build_spec(entity, kind, ignore_validation=True, module_kind=kind, **obj.get("spec"))
        status = build_status(entity, **obj.get("status"))
        return {
            "uuid": uuid,
            "kind": kind,
            "metadata": metadata,
            "spec": spec,
            "status": status,
        }


def function_from_parameters(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    embedded: bool = True,
    source: str | None = None,
    **kwargs,
) -> Function:
    """
    Create a new Function instance and persist it to the backend.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        Identifier of the function.
    kind : str
        The type of the function.
    uuid : str
        UUID.
    description : str
        Description of the function.
    embedded : bool
        Flag to determine if object must be embedded in project.
    source : str
        Path to the function's source code on the local file system.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Function
       Object instance.
    """
    uuid = build_uuid(uuid)
    spec = build_spec(
        FUNC,
        kind,
        module_kind=kind,
        source=source,
        **kwargs,
    )
    metadata = build_metadata(
        FUNC,
        project=project,
        name=name,
        version=uuid,
        description=description,
        embedded=embedded,
    )
    status = build_status(FUNC)
    return Function(
        uuid=uuid,
        kind=kind,
        metadata=metadata,
        spec=spec,
        status=status,
    )


def function_from_dict(obj: dict) -> Function:
    """
    Create function from dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create function from.

    Returns
    -------
    Function
        Function object.
    """
    return Function.from_dict(FUNC, obj)



"""
Artifact base specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.spec import Spec
from pydantic import BaseModel


class ArtifactSpec(Spec):
    """
    Artifact specification.
    """

    def __init__(
        self,
        key: str | None = None,
        src_path: str | None = None,
        target_path: str | None = None,
        **kwargs,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        key : str
            The key of the artifact.
        src_path : str
            The source path of the artifact.
        target_path : str
            The target path of the artifact.
        **kwargs
            Keywords arguments.
        """
        self.key = key
        self.src_path = src_path
        self.target_path = target_path

        self._any_setter(**kwargs)


class ArtifactParams(BaseModel):
    """
    Artifact base parameters.
    """

    key: str = None
    """Key of the artifact"""

    src_path: str = None
    """Source path of the artifact."""

    target_path: str = None
    """Target path of the artifact."""



"""
Workflow module.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities._base.entity import Entity
from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.utils.api import api_ctx_create, api_ctx_update
from digitalhub_core.utils.commons import WKFL
from digitalhub_core.utils.generic_utils import build_uuid, get_timestamp
from digitalhub_core.utils.io_utils import write_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.context.context import Context
    from digitalhub_core.entities.workflows.metadata import WorkflowMetadata
    from digitalhub_core.entities.workflows.spec import WorkflowSpec
    from digitalhub_core.entities.workflows.status import WorkflowStatus


class Workflow(Entity):
    """
    A class representing a workflow.
    """

    def __init__(
        self,
        uuid: str,
        kind: str,
        metadata: WorkflowMetadata,
        spec: WorkflowSpec,
        status: WorkflowStatus,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        uuid : str
            UUID.
        kind : str
            Kind of the object.
        metadata : WorkflowMetadata
            Metadata of the object.
        spec : WorkflowSpec
            Specification of the object.
        status : WorkflowStatus
            Status of the object.
        """
        super().__init__()
        self.id = uuid
        self.kind = kind
        self.metadata = metadata
        self.spec = spec
        self.status = status

        self.project = self.metadata.project
        self.name = self.metadata.name
        self.embedded = self.metadata.embedded
        self._obj_attr.extend(["project", "name", "embedded"])

    #############################
    #  Save / Export
    #############################

    def save(self, update: bool = False) -> dict:
        """
        Save workflow into backend.

        Parameters
        ----------
        update : bool
            Flag to indicate update.

        Returns
        -------
        dict
            Mapping representation of Workflow from backend.
        """
        obj = self.to_dict()

        if not update:
            api = api_ctx_create(self.metadata.project, WKFL)
            return self._context().create_object(obj, api)

        self.metadata.updated = obj["metadata"]["updated"] = get_timestamp()
        api = api_ctx_update(self.metadata.project, WKFL, self.metadata.name, self.id)
        return self._context().update_object(obj, api)

    def export(self, filename: str | None = None) -> None:
        """
        Export object as a YAML file.

        Parameters
        ----------
        filename : str
            Name of the export YAML file. If not specified, the default value is used.

        Returns
        -------
        None
        """
        obj = self.to_dict()
        filename = filename if filename is not None else f"workflow_{self.metadata.project}_{self.metadata.name}.yaml"
        write_yaml(filename, obj)

    #############################
    #  Context
    #############################

    def _context(self) -> Context:
        """
        Get context.

        Returns
        -------
        Context
            Context.
        """
        return get_context(self.metadata.project)


def workflow_from_parameters(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    embedded: bool = True,
    **kwargs,
) -> Workflow:
    """
    Create a new Workflow instance with the specified parameters.

    Parameters
    ----------
    project : str
        A string representing the project associated with this workflow.
    name : str
        The name of the workflow.
    kind : str
        Kind of the object.
    uuid : str
        UUID.
    description : str
        A description of the workflow.
    embedded : bool
        Flag to determine if object must be embedded in project.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Workflow
        An instance of the created workflow.
    """
    uuid = build_uuid(uuid)
    metadata = build_metadata(
        WKFL,
        project=project,
        name=name,
        version=uuid,
        description=description,
        embedded=embedded,
    )
    spec = build_spec(
        WKFL,
        kind,
        **kwargs,
    )
    status = build_status(WKFL)
    return Workflow(
        uuid=uuid,
        kind=kind,
        metadata=metadata,
        spec=spec,
        status=status,
    )


def workflow_from_dict(obj: dict) -> Workflow:
    """
    Create Workflow instance from a dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create Workflow from.

    Returns
    -------
    Workflow
        Workflow instance.
    """
    return Workflow.from_dict(WKFL, obj)



"""
Status class module.
"""
from __future__ import annotations

from enum import Enum

from digitalhub_core.entities._base.base import ModelObj


class State(Enum):
    """
    State enumeration.
    """

    BUILT = "BUILT"
    COMPLETED = "COMPLETED"
    CREATED = "CREATED"
    ERROR = "ERROR"
    IDLE = "IDLE"
    PENDING = "PENDING"
    READY = "READY"
    RUNNING = "RUNNING"
    STOP = "STOP"


class Status(ModelObj):
    """
    Base Status class.
    """

    def __init__(self, state: str, message: str | None = None) -> None:
        """
        Constructor.

        Parameters
        ----------
        state : str
            The state of the entity.
        message : str
            Error message.
        """
        self.state = state
        self.message = message

    @classmethod
    def from_dict(cls, obj: dict) -> "Status":
        """
        Return entity specification object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity specification.

        Returns
        -------
        EntitySpec
            An entity specification object.
        """
        return cls(**obj)



"""
Import modules from submodules.
"""
from digitalhub_core.entities.artifacts.crud import (
    delete_artifact,
    get_artifact,
    import_artifact,
    new_artifact,
    update_artifact,
)
from digitalhub_core.entities.dataitems.crud import (
    delete_dataitem,
    get_dataitem,
    import_dataitem,
    new_dataitem,
    update_dataitem,
)
from digitalhub_core.entities.functions.crud import (
    delete_function,
    get_function,
    import_function,
    new_function,
    update_function,
)
from digitalhub_core.entities.projects.crud import (
    delete_project,
    get_or_create_project,
    get_project,
    import_project,
    new_project,
    update_project,
)
from digitalhub_core.entities.runs.crud import delete_run, get_run, import_run, new_run, update_run
from digitalhub_core.entities.tasks.crud import delete_task, get_task, import_task, new_task, update_task
from digitalhub_core.entities.workflows.crud import (
    delete_workflow,
    get_workflow,
    import_workflow,
    new_workflow,
    update_workflow,
)
from digitalhub_core.stores.builder import set_store
from digitalhub_core.utils.generic_utils import set_dhub_env



"""
Run base specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.spec import Spec
from pydantic import BaseModel


class RunSpec(Spec):
    """Run specification."""

    def __init__(
        self,
        task: str,
        task_id: str,
        inputs: dict | None = None,
        outputs: dict | None = None,
        parameters: dict | None = None,
        local_execution: bool = False,
        **kwargs,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        task : str
            The task associated with the run.
        task_id : str
            The task id associated with the run.
        inputs : dict
            The inputs of the run.
        outputs : dict
            The outputs of the run.
        parameters : dict
            The parameters of the run.
        local_execution : bool
            Flag to indicate if the run will be executed locally
        **kwargs
            Keywords arguments.
        """
        self.task = task
        self.task_id = task_id
        self.inputs = inputs if inputs is not None else {}
        self.outputs = outputs if outputs is not None else {}
        self.parameters = parameters if parameters is not None else {}
        self.local_execution = local_execution

        self._any_setter(**kwargs)


class Objects(BaseModel):
    """
    Run inputs specification.
    """

    dataitems: list[str] = None
    """List of dataitems names."""

    artifacts: list[str] = None
    """List of artifacts names."""


class RunParams(BaseModel):
    """
    Run parameters.
    """

    task: str = None
    """The task string associated with the run."""

    task_id: str = None
    """The task id associated with the run."""

    inputs: Objects = None
    """List of input dataitems and artifacts names."""

    outputs: Objects = None
    """List of output dataitems and artifacts names."""

    parameters: dict = None
    """Parameters to be used in the run."""

    local_execution: bool = False
    """Flag to indicate if the run will be executed locally."""



"""
Task module.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities._base.entity import Entity
from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.entities.runs.crud import delete_run, get_run, new_run
from digitalhub_core.utils.api import api_base_create, api_base_update
from digitalhub_core.utils.commons import TASK
from digitalhub_core.utils.generic_utils import build_uuid, get_timestamp
from digitalhub_core.utils.io_utils import write_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.context.context import Context
    from digitalhub_core.entities.runs.entity import Run
    from digitalhub_core.entities.tasks.metadata import TaskMetadata
    from digitalhub_core.entities.tasks.spec import TaskSpec
    from digitalhub_core.entities.tasks.status import TaskStatus


class Task(Entity):
    """
    A class representing a task.
    """

    def __init__(
        self,
        uuid: str,
        kind: str,
        metadata: TaskMetadata,
        spec: TaskSpec,
        status: TaskStatus,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        uuid : str
            UUID.
        kind : str
            Kind of the object.
        metadata : TaskMetadata
            Metadata of the object.
        spec : TaskSpec
            Specification of the object.
        status : TaskStatus
            Status of the object.
        """
        super().__init__()
        self.id = uuid
        self.kind = kind
        self.metadata = metadata
        self.spec = spec
        self.status = status

        self.project = self.metadata.project
        self._obj_attr.extend(["project"])

    #############################
    #  Save / Export
    #############################

    def save(self, update: bool = False) -> dict:
        """
        Save task into backend.

        Parameters
        ----------
        update : bool
            Flag to indicate update.

        Returns
        -------
        dict
            Mapping representation of Task from backend.
        """
        obj = self.to_dict()

        if not update:
            api = api_base_create(TASK)
            return self._context().create_object(obj, api)

        self.metadata.updated = obj["metadata"]["updated"] = get_timestamp()
        api = api_base_update(TASK, self.id)
        return self._context().update_object(obj, api)

    def export(self, filename: str | None = None) -> None:
        """
        Export object as a YAML file.

        Parameters
        ----------
        filename : str
            Name of the export YAML file. If not specified, the default value is used.

        Returns
        -------
        None
        """
        obj = self.to_dict()
        filename = filename if filename is not None else f"task_{self.id}.yaml"
        write_yaml(filename, obj)

    #############################
    #  Context
    #############################

    def _context(self) -> Context:
        """
        Get context.

        Returns
        -------
        Context
            Context.
        """
        return get_context(self.metadata.project)

    #############################
    #  Task methods
    #############################

    def run(
        self,
        inputs: dict | None,
        outputs: dict | None = None,
        parameters: dict | None = None,
        local_execution: bool = False,
    ) -> Run:
        """
        Run task.

        Parameters
        ----------
        inputs : dict
            The inputs of the run.
        outputs : dict
            The outputs of the run.
        parameters : dict
            The parameters of the run.
        local_execution : bool
            Flag to indicate if the run will be executed locally.

        Returns
        -------
        Run
            Run object.
        """
        return self.new_run(
            project=self.metadata.project,
            task=self._get_task_string(),
            task_id=self.id,
            kind="run",
            inputs=inputs,
            outputs=outputs,
            parameters=parameters,
            local_execution=local_execution,
        )

    def _get_task_string(self) -> str:
        """
        Get task string.

        Returns
        -------
        str
            Task string.
        """
        splitted = self.spec.function.split("://")
        return f"{splitted[0]}+{self.kind}://{splitted[1]}"

    #############################
    # CRUD Methods for Run
    #############################

    def new_run(self, **kwargs) -> Run:
        """
        Create a new run.

        Parameters
        ----------
        **kwargs
            Keyword arguments.

        Returns
        -------
        Run
            Run object.
        """
        return new_run(**kwargs)

    def get_run(self, uuid: str) -> Run:
        """
        Get run.

        Parameters
        ----------
        update : bool
            Flag to indicate update.

        Returns
        -------
        Run
            Run object.
        """
        return get_run(self.metadata.project, uuid)

    def delete_run(self, uuid: str) -> None:
        """
        Delete run.

        Parameters
        ----------
        update : bool
            Flag to indicate update.

        Returns
        -------
        None
        """
        delete_run(self.metadata.project, uuid)

    #############################
    #  Overridden Methods
    #############################

    @staticmethod
    def _parse_dict(entity: str, obj: dict) -> dict:
        """
        Get dictionary and parse it to a valid entity dictionary.

        Parameters
        ----------
        entity : str
            Entity type.
        obj : dict
            Dictionary to parse.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        uuid = build_uuid(obj.get("id"))
        kind = obj.get("kind", "")
        metadata = build_metadata(entity, **obj.get("metadata"))
        spec = build_spec(entity, kind, ignore_validation=True, module_kind=kind, **obj.get("spec"))
        status = build_status(entity, **obj.get("status"))
        return {
            "uuid": uuid,
            "kind": kind,
            "metadata": metadata,
            "spec": spec,
            "status": status,
        }


def task_from_parameters(
    project: str,
    kind: str,
    uuid: str | None = None,
    function: str | None = "",
    **kwargs,
) -> Task:
    """
    Create a new object instance.

    Parameters
    ----------
    project : str
        Name of the project.
    kind : str
        The type of the task.
    uuid : str
        UUID.
    function : str
        The function string identifying the function.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Task
       Object instance.
    """
    uuid = build_uuid(uuid)
    metadata = build_metadata(TASK, project=project, name=uuid)
    spec = build_spec(
        TASK,
        kind,
        module_kind=function.split("://")[0],
        function=function,
        **kwargs,
    )
    status = build_status(TASK)
    return Task(
        uuid=uuid,
        kind=kind,
        metadata=metadata,
        spec=spec,
        status=status,
    )


def task_from_dict(obj: dict) -> Task:
    """
    Create Task object from dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary representation of Task.

    Returns
    -------
    Task
        Task object.
    """
    return Task.from_dict(TASK, obj)



"""
S3Store module.
"""
from __future__ import annotations

import re

import pandas as pd
from digitalhub_core.stores.objects.base import Store, StoreConfig
from digitalhub_core.utils.exceptions import StoreError
from digitalhub_core.utils.file_utils import build_path
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.exc import SQLAlchemyError


class SQLStoreConfig(StoreConfig):
    """
    SQL store configuration class.
    """

    host: str
    """SQL host."""

    port: int
    """SQL port."""

    user: str
    """SQL user."""

    password: str
    """SQL password."""

    database: str
    """SQL database name."""

    pg_schema: str
    """SQL schema name."""


class SqlStore(Store):
    """
    SQL store class. It implements the Store interface and provides methods to fetch and persist
    artifacts on SQL based storage.
    """

    def __init__(self, name: str, store_type: str, config: SQLStoreConfig) -> None:
        """
        Constructor.

        Parameters
        ----------
        config : SQLStoreConfig
            SQL store configuration.

        See Also
        --------
        Store.__init__
        """
        super().__init__(name, store_type)
        self.config = config

    ############################
    # IO methods
    ############################

    def download(self, src: str, dst: str | None = None) -> str:
        """
        Download an artifact from SQL based storage.

        See Also
        --------
        fetch_artifact
        """
        return self._registry.get(src, self.fetch_artifact(src, dst))

    def fetch_artifact(self, src: str, dst: str | None = None) -> str:
        """
        Fetch an artifact from SQL based storage. If the destination is not provided,
        a temporary directory will be created and the artifact will be saved there.

        Parameters
        ----------
        src : str
            Table name.
        dst : str
            The destination of the artifact on local filesystem.

        Returns
        -------
        str
            Returns a file path.
        """
        dst = dst if dst is not None else self._build_temp(src)
        dst = build_path(dst, "data.parquet")
        schema = self._get_schema(src)
        table = self._get_table_name(src)
        return self._download_table(schema, table, dst)

    def upload(self, src: str, dst: str | None = None) -> str:
        """
        Method to upload an artifact to the backend. Please note that this method is not implemented
        since the SQL store is not meant to upload artifacts.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError("SQL store does not support upload.")

    def persist_artifact(self, src: str, dst: str | None = None) -> str:
        """
        Method to persist an artifact. Note that this method is not implemented
        since the SQL store is not meant to write artifacts.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError("SQL store does not support persist_artifact.")

    def write_df(self, df: pd.DataFrame, dst: str | None = None, **kwargs) -> str:
        """
        Write a dataframe to a database. Kwargs are passed to df.to_sql().

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe.
        dst : str
            The destination table on database.
        **kwargs
            Keyword arguments.

        Returns
        -------
        str
            The SQL uri where the dataframe was saved.
        """
        if dst is None:
            schema = str(self.config.pg_schema)
            table = "table"
        else:
            schema = self._get_schema(dst)
            table = self._get_table_name(dst)
        return self._upload_table(df, schema, table, **kwargs)

    ############################
    # Private helper methods
    ############################

    def _get_connection_string(self) -> str:
        """
        Get the connection string.

        Returns
        -------
        str
            The connection string.
        """
        return (
            f"postgresql+psycopg2://{self.config.user}:{self.config.password}@"
            f"{self.config.host}:{self.config.port}/{self.config.database}"
        )

    def _get_engine(self) -> Engine:
        """
        Create engine from connection string.

        Returns
        -------
        Engine
            An SQLAlchemy engine.
        """
        connection_string = self._get_connection_string()
        if not isinstance(connection_string, str):
            raise StoreError("Connection string must be a string.")
        try:
            return create_engine(connection_string)
        except Exception as ex:
            raise StoreError(f"Something wrong with connection string. Arguments: {str(ex.args)}")

    def _check_factory(self) -> Engine:
        """
        Check if the database is accessible and return the engine.

        Returns
        -------
        Engine
            The database engine.
        """
        engine = self._get_engine()
        self._check_access_to_storage(engine)
        return engine

    @staticmethod
    def _parse_path(path: str) -> dict:
        """
        Parse the path and return the components.

        Parameters
        ----------
        path : str
            The path.

        Returns
        -------
        dict
            A dictionary containing the components of the path.
        """
        pattern = r"^sql:\/\/(postgres\/)?(?P<database>.+)?\/(?P<schema>.+)\/(?P<table>.+)$"
        match = re.match(pattern, path)
        if match is None:
            raise ValueError("Invalid SQL path. Must be sql://postgres/<database>/<schema>/<table>")
        return match.groupdict()

    def _get_schema(self, uri: str) -> str:
        """
        Get the name of the SQL schema from the URI.

        Parameters
        ----------
        uri : str
            The URI.

        Returns
        -------
        str
            The name of the SQL schema.
        """
        return str(self._parse_path(uri).get("schema"))

    def _get_table_name(self, uri: str) -> str:
        """
        Get the name of the table from the URI.

        Parameters
        ----------
        uri : str
            The URI.

        Returns
        -------
        str
            The name of the table
        """
        return str(self._parse_path(uri).get("table"))

    @staticmethod
    def _check_access_to_storage(engine: Engine) -> None:
        """
        Check if there is access to the storage.

        Parameters
        ----------
        engine : Engine
            An SQLAlchemy engine.

        Returns
        -------
        None

        Raises
        ------
        StoreError
            If there is no access to the storage.
        """
        try:
            engine.connect()
        except SQLAlchemyError:
            engine.dispose()
            raise StoreError("No access to db!")

    def _download_table(self, schema: str, table: str, dst: str) -> str:
        """
        Download a table from SQL based storage.

        Parameters
        ----------
        schema : str
            The origin schema.
        table : str
            The origin table.
        dst : str
            The destination path.

        Returns
        -------
        str
            The destination path.
        """
        engine = self._check_factory()
        self._check_local_dst(dst)
        pd.read_sql_table(table, engine, schema=schema).to_parquet(dst, index=False)
        engine.dispose()
        return dst

    def _upload_table(self, df: pd.DataFrame, schema: str, table: str, **kwargs) -> str:
        """
        Upload a table to SQL based storage.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe.
        schema : str
            Destination schema.
        table : str
            Destination table.
        **kwargs
            Keyword arguments.

        Returns
        -------
        str
            The SQL URI where the dataframe was saved.
        """
        engine = self._check_factory()
        df.to_sql(table, engine, schema=schema, index=False, **kwargs)
        engine.dispose()
        return f"sql://postgres/{engine.url.database}/{schema}/{table}"

    ############################
    # Store interface methods
    ############################

    @staticmethod
    def is_local() -> bool:
        """
        Check if the store is local.

        Returns
        -------
        bool
            False
        """
        return False



"""
Project base specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.spec import Spec
from pydantic import BaseModel


class ProjectSpec(Spec):
    """
    Project specification.
    """

    def __init__(
        self,
        context: str | None = None,
        source: str | None = None,
        functions: list | None = None,
        artifacts: list | None = None,
        workflows: list | None = None,
        dataitems: list | None = None,
    ) -> None:
        """
        Initializes a new instance of the ProjectSpec class.

        Parameters
        ----------
        context : str
            The project's context.

        source : str
            The project's source.

        functions : list
            List of project's functions.

        artifacts : list
            List of project's artifacts.

        workflows : list
            List of project's workflows.

        dataitems : list
            List of project's dataitems.

        Returns
        -------
        None
        """
        self.context = context
        self.source = source
        self.functions = functions if functions is not None else []
        self.artifacts = artifacts if artifacts is not None else []
        self.workflows = workflows if workflows is not None else []
        self.dataitems = dataitems if dataitems is not None else []


class ProjectParams(BaseModel):
    """
    Parameters model for project.
    """

    context: str = None
    """The project's context."""

    source: str = None
    """The project's source."""

    functions: list = None
    """List of project's functions."""

    artifacts: list = None
    """List of project's artifacts."""

    workflows: list = None
    """List of project's workflows."""

    dataitems: list = None
    """List of project's dataitems."""



"""
APIs module.
"""
from __future__ import annotations

API_BASE = "/api/v1"
API_CONTEXT = f"{API_BASE}/-"

####################
# Context APIs
####################


def api_ctx_create(
    proj: str,
    dto: str,
) -> str:
    """
    Create context API.

    Parameters
    ----------
    proj : str
        Name of the project.
    dto : str
        The name of the DTO.

    Returns
    -------
    str
        The API string formatted.
    """
    return f"{API_CONTEXT}/{proj}/{dto}"


def api_ctx_read(
    proj: str,
    dto: str,
    name: str,
    uuid: str | None = None,
) -> str:
    """
    Read context API.

    Parameters
    ----------
    proj : str
        Name of the project.
    dto : str
        The type of the DTO.
    name : str
        The name of the DTO.
    uuid : str
        The UUID of the DTO. If not provided, the latest version is returned.

    Returns
    -------
    str
        The API string formatted.
    """
    version = f"/{uuid}" if uuid is not None else "/latest"
    return f"{API_CONTEXT}/{proj}/{dto}/{name}{version}"


def api_ctx_update(
    proj: str,
    dto: str,
    name: str,
    uuid: str,
) -> str:
    """
    Update context API.

    Parameters
    ----------
    proj : str
        Name of the project.
    dto : str
        The type of the DTO.
    name : str
        The name of the DTO.
    uuid : str
        The UUID of the DTO.

    Returns
    -------
    str
        The API string formatted.
    """
    return f"{API_CONTEXT}/{proj}/{dto}/{name}/{uuid}"


def api_ctx_delete(
    proj: str,
    dto: str,
    name: str,
    uuid: str | None = None,
) -> str:
    """
    Delete context API.

    Parameters
    ----------
    proj : str
        Name of the project.
    dto : str
        The type of the DTO.
    name : str
        The name of the DTO.
    uuid : str
        The UUID of the DTO. If not provided, all versions are deleted.

    Returns
    -------
    str
        The API string formatted.
    """
    version = f"/{uuid}" if uuid is not None else ""
    return f"{API_CONTEXT}/{proj}/{dto}/{name}{version}"


####################
# Base controller APIs
####################


def api_base_create(dto: str) -> str:
    """
    Create base API.

    Parameters
    ----------
    dto : str
        The type of the DTO.

    Returns
    -------
    str
        The API string formatted.
    """
    return f"{API_BASE}/{dto}"


def api_base_read(dto: str, name: str) -> str:
    """
    Read base API.

    Parameters
    ----------
    dto : str
        The type of the DTO.
    name : str
        The name or UUID of the DTO.

    Returns
    -------
    str
        The API string formatted.
    """
    return f"{API_BASE}/{dto}/{name}"


def api_base_update(dto: str, name: str) -> str:
    """
    Update base API.

    Parameters
    ----------
    dto : str
        The type of the DTO.
    name : str
        The name or UUID of the DTO.

    Returns
    -------
    str
        The API string formatted.
    """
    return f"{API_BASE}/{dto}/{name}"


def api_base_delete(dto: str, name: str) -> str:
    """
    Delete base API.

    Parameters
    ----------
    dto : str
        The type of the DTO.
    name : str
        The name or UUID of the DTO.

    Returns
    -------
    str
        The API string formatted.
    """
    return f"{API_BASE}/{dto}/{name}"



"""
Project metadata module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.metadata import Metadata


class ProjectMetadata(Metadata):
    """
    A class representing Project metadata.
    """

    def __init__(
        self,
        project: str,
        created: str | None = None,
        updated: str | None = None,
        name: str | None = None,
        version: str | None = None,
        description: str | None = None,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        name : str
            Name of the object.
        version : str
            Version of the object.
        description : str
            Description of the entity.

        See Also
        --------
        Metadata.__init__
        """
        super().__init__(project, created, updated)
        self.name = name
        self.version = version
        self.description = description



"""
Module for performing operations on Dataitem objects.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities.dataitems.entity import dataitem_from_dict, dataitem_from_parameters
from digitalhub_core.utils.api import api_ctx_delete, api_ctx_read, api_ctx_update
from digitalhub_core.utils.commons import DTIT
from digitalhub_core.utils.entities_utils import parse_entity_key
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.dataitems.entity import Dataitem


def create_dataitem(**kwargs) -> Dataitem:
    """
    Create a new data item with the provided parameters.

    Parameters
    ----------
    **kwargs
        Keyword arguments.

    Returns
    -------
    Dataitem
        Object instance.
    """
    return dataitem_from_parameters(**kwargs)


def create_dataitem_from_dict(obj: dict) -> Dataitem:
    """
    Create a new Dataitem instance from a dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create the Dataitem from.

    Returns
    -------
    Dataitem
        Dataitem object.
    """
    return dataitem_from_dict(obj)


def new_dataitem(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    embedded: bool = True,
    key: str | None = None,
    path: str | None = None,
    **kwargs,
) -> Dataitem:
    """
    Create a new object instance.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        Identifier of the dataitem.
    kind : str
        The type of the dataitem.
    uuid : str
        UUID.
    description : str
        Description of the dataitem.
    embedded : bool
        Flag to determine if object must be embedded in project.
    key : str
        Representation of the dataitem, e.g. store://etc.
    path : str
        Path to the dataitem on local file system or remote storage.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Dataitem
       Object instance.
    """
    obj = create_dataitem(
        project=project,
        name=name,
        kind=kind,
        uuid=uuid,
        description=description,
        embedded=embedded,
        key=key,
        path=path,
        **kwargs,
    )
    obj.save()
    return obj


def get_dataitem(project: str, name: str, uuid: str | None = None) -> Dataitem:
    """
    Get object from backend.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the dataitem.
    uuid : str
        UUID.

    Returns
    -------
    Dataitem
        Object instance.

    """
    api = api_ctx_read(project, DTIT, name, uuid=uuid)
    obj = get_context(project).read_object(api)
    return dataitem_from_dict(obj)


def get_dataitem_from_key(key: str) -> Dataitem:
    """
    Get dataitem from key.

    Parameters
    ----------
    key : str
        Key of the dataitem.
        It's format is store://<project>/dataitems/<kind>/<name>:<uuid>.
    """
    project, name, uuid = parse_entity_key(key)
    return get_dataitem(project, name, uuid)


def import_dataitem(file: str) -> Dataitem:
    """
    Get object from file.

    Parameters
    ----------
    file : str
        Path to the file.

    Returns
    -------
    Dataitem
        Object instance.
    """
    obj = read_yaml(file)
    return dataitem_from_dict(obj)


def delete_dataitem(project: str, name: str, uuid: str | None = None) -> dict:
    """
    Delete dataitem from the backend. If the uuid is not specified, delete all versions.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the dataitem.
    uuid : str
        UUID.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_ctx_delete(project, DTIT, name, uuid=uuid)
    return get_context(project).delete_object(api)


def update_dataitem(dataitem: Dataitem) -> dict:
    """
    Update a dataitem.

    Parameters
    ----------
    dataitem : Dataitem
        The dataitem to update.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_ctx_update(dataitem.metadata.project, DTIT, dataitem.metadata.name, uuid=dataitem.id)
    return get_context(dataitem.metadata.project).update_object(dataitem.to_dict(), api)



"""
Local store module.
"""
from __future__ import annotations

import typing

from digitalhub_core.stores.objects.base import Store, StoreConfig
from digitalhub_core.utils.file_utils import build_path, copy_file, get_dir
from digitalhub_core.utils.uri_utils import get_name_from_uri, get_uri_path

if typing.TYPE_CHECKING:
    import pandas as pd


class LocalStoreConfig(StoreConfig):
    """
    Local store configuration class.
    """

    path: str
    """Local path."""


class LocalStore(Store):
    """
    Local store class. It implements the Store interface and provides methods to fetch and persist
    artifacts on local filesystem based storage.
    """

    def __init__(self, name: str, store_type: str, config: LocalStoreConfig) -> None:
        """
        Constructor.

        Parameters
        ----------
        config : LocalStoreConfig
            Local store configuration.

        See Also
        --------
        Store.__init__
        """
        super().__init__(name, store_type)
        self.config = config

    ############################
    # IO methods
    ############################

    def download(self, src: str, dst: str | None = None) -> str:
        """
        Method to download an artifact from the backend. Please note that this method is not
        implemented since the local store is not meant to download artifacts.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError("Local store does not support download. Use as_file() instead.")

    def fetch_artifact(self, src: str, dst: str | None = None) -> str:
        """
        Method to fetch an artifact from the backend and to register it on the paths registry.
        If destination is not provided, return the source path, otherwise the path of the copied
        file.

        Parameters
        ----------
        src : str
            The source location of the artifact.
        dst : str
            The destination of the artifact.

        Returns
        -------
        str
            Returns the path of the artifact.
        """
        if dst is None:
            return src
        self._check_local_dst(dst)
        return copy_file(src, dst)

    def upload(self, src: str, dst: str | None = None) -> str:
        """
        Method to upload an artifact to the backend. Please note that this method is not implemented
        since the local store is not meant to upload artifacts.

        Returns
        -------
        None

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError("Local store does not support upload.")

    def persist_artifact(self, src: str, dst: str | None = None) -> str:
        """
        Method to persist (copy) an artifact on local filesystem.
        If destination is not provided, the artifact is written to the default
        store path with source name.

        Parameters
        ----------
        src : str
            The source location of the artifact.
        dst : str
            The destination of the artifact.
        src_name : str
            The name of the artifact.

        Returns
        -------
        dst
            Returns the URI of the artifact.
        """
        if dst is None:
            dst = build_path(get_dir(self.config.path), get_name_from_uri(src))
        self._check_local_dst(dst)
        return copy_file(src, dst)

    def write_df(self, df: pd.DataFrame, dst: str | None = None, **kwargs) -> str:
        """
        Method to write a dataframe to a file. Kwargs are passed to df.to_parquet().
        If destination is not provided, the dataframe is written to the default
        store path with name data.parquet.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to write.
        dst : str
            The destination of the dataframe.
        **kwargs
            Keyword arguments.

        Returns
        -------
        str
            Path of written dataframe.
        """
        if dst is None or not dst.endswith(".parquet"):
            dst = build_path(get_uri_path(self.config.path), "data.parquet")
        self._check_local_dst(dst)
        df.to_parquet(dst, index=False, **kwargs)
        return dst

    ############################
    # Store interface methods
    ############################

    @staticmethod
    def is_local() -> bool:
        """
        Check if the store is local.

        Returns
        -------
        bool
            True
        """
        return True



"""
Module to organize the import of extension modules.
"""
from __future__ import annotations

import importlib
from collections import namedtuple

from digitalhub_core.utils.commons import FUNC

Rtm = namedtuple("Rtm", ["module", "class_name"])
Obj = namedtuple("Obj", ["module", "class_spec", "class_params"])


class ModuleRegistry:
    """
    Registry for modules to organize the import of classes.

    Attributes
    ----------
    runtime : Rtm
        Runtime class.
    function : Obj
        Function class.
    tasks : dict[str, Obj]
        Tasks classes.

    Methods
    -------
    register_runtime
        Register a runtime class.
    register_function
        Register a function class.
    register_tasks
        Register a task class.
    get_runtime
        Get runtime class.
    get_spec
        Get function or task spec class and validators.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """
        self.runtime: Rtm | None = None
        self.function: Obj | None = None
        self.tasks: dict[str, Obj] | None = {}

    def register_runtime(self, module: str, class_name: str) -> None:
        """
        Register a runtime class.

        Parameters
        ----------
        module : str
            Module name.
        class_name : str
            Class name.

        Returns
        -------
        None
        """
        self.runtime = Rtm(module, class_name)

    def register_function(self, module: str, class_spec: str, class_params: str) -> None:
        """
        Register a function class.

        Parameters
        ----------
        module : str
            Module name.
        class_spec : str
            Class spec.
        class_params : str
            Class params.

        Returns
        -------
        None
        """
        self.function = Obj(module, class_spec, class_params)

    def register_tasks(self, kind: str, module: str, class_spec: str, class_params: str) -> None:
        """
        Register a task class.

        Parameters
        ----------
        kind : str
            Task kind.
        module : str
            Module name.
        class_spec : str
            Class spec.
        class_params : str
            Class params.

        Returns
        -------
        None
        """
        self.tasks[kind] = Obj(module, class_spec, class_params)

    def get_runtime(self) -> Obj:
        """
        Get runtime class.

        Returns
        -------
        Obj
            Runtime class.
        """
        return self.runtime

    def get_spec(self, entity: str, kind: str | None = None) -> Obj:
        """
        Get function or task spec class and validators.

        Parameters
        ----------
        entity : str
            Entity type.
        kind : str
            Entity kind.

        Returns
        -------
        Obj
            Spec class.
        """
        if entity == FUNC:
            return self.function
        return self.tasks[kind]


def import_registry(kind: str) -> ModuleRegistry:
    """
    Import registry from implemented module.

    Parameters
    ----------
    kind : str
        Module kind. Basically, the kind of the function due to correspondence
        function-runtime name.

    Returns
    -------
    ModuleRegistry
        Registry.
    """
    try:
        module = importlib.import_module(f"digitalhub_core_{kind}")
        return getattr(module, "registry")
    except (ModuleNotFoundError, ImportError):
        raise ValueError(f"Module digitalhub_core_{kind} not found")



"""
Task specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.spec import Spec
from pydantic import BaseModel


class TaskSpec(Spec):
    """Task specification."""

    def __init__(self, function: str | None = None, **kwargs) -> None:
        """
        Constructor.

        Parameters
        ----------
        function : str
            Function string.
        **kwargs
            Keyword arguments.
        """
        self.function = function

        self._any_setter(**kwargs)


class TaskParams(BaseModel):
    """
    Base task model.
    """

    function: str = None
    """Task function."""



"""
RunStatus class module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.status import Status


class RunStatus(Status):
    """
    Status class for run entities.
    """

    def __init__(
        self,
        state: str | None = None,
        dataitems: list | None = None,
        artifacts: list | None = None,
        timing: dict | None = None,
        **kwargs,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        dataitems : list
            Dataitems infos.
        artifacts : list
            Artifacts infos.
        timings : dict
            Run execution times.
        **kwargs
            Keyword arguments.


        See Also
        --------
        Status.__init__
        """
        super().__init__(state, **kwargs)
        self.dataitems = dataitems
        self.artifacts = artifacts
        self.timing = timing



"""
Logger module.
"""
from __future__ import annotations

import logging

# LOGGER
LOGGER = logging.getLogger("digitalhub-core")
LOGGER.setLevel(logging.INFO)


# Create console handler set level to INFO and add filter
class LogFilter(logging.Filter):
    """
    Logs filter.
    """

    def filter(self, record):
        """
        Filter for nefertem logs.
        """
        return record.name == "digitalhub-core"


ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
ch.addFilter(LogFilter())

# Create formatter and add to console handler
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)

# Add console handler to logger
LOGGER.addHandler(ch)



"""
Metadata factory module.
"""
from __future__ import annotations

import typing

from digitalhub_core.entities.artifacts.metadata import ArtifactMetadata
from digitalhub_core.entities.dataitems.metadata import DataitemMetadata
from digitalhub_core.entities.functions.metadata import FunctionMetadata
from digitalhub_core.entities.projects.metadata import ProjectMetadata
from digitalhub_core.entities.runs.metadata import RunMetadata
from digitalhub_core.entities.tasks.metadata import TaskMetadata
from digitalhub_core.entities.workflows.metadata import WorkflowMetadata
from digitalhub_core.utils.commons import ARTF, DTIT, FUNC, PROJ, RUNS, TASK, WKFL
from digitalhub_core.utils.generic_utils import get_timestamp

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.metadata import Metadata


class MetadataBuilder(dict):
    """
    Metadata builder class.
    """

    def register(self, module: str, metadata: Metadata) -> None:
        """
        Register metadata.

        Parameters
        ----------
        module: str
            module name.
        metadata: Metadata
            Metadata object.

        Returns
        -------
        None
        """
        self[module] = metadata

    def build(self, module: str, **kwargs) -> Metadata:
        """
        Build entity metadata object.

        Parameters
        ----------
        module: str
            module name.
        **kwargs
            Keyword arguments.

        Returns
        -------
        Metadata
            An entity metadata object.
        """
        if module not in self:
            raise ValueError(f"Invalid module name: {module}")
        kwargs = self._parse_arguments(**kwargs)
        return self[module](**kwargs)

    @staticmethod
    def _parse_arguments(**kwargs) -> dict:
        """
        Parse keyword arguments and add default values.

        Parameters
        ----------
        **kwargs
            Keyword arguments.

        Returns
        -------
        dict
            A dictionary containing the entity metadata attributes.
        """
        if "created" not in kwargs or kwargs["created"] is None:
            kwargs["created"] = get_timestamp()
        if "updated" not in kwargs or kwargs["updated"] is None:
            kwargs["updated"] = kwargs["created"]
        return kwargs


def build_metadata(module: str, **kwargs) -> Metadata:
    """
    Wrapper for MetadataBuilder.build_metadata.

    Parameters
    ----------
    module: str
        Module name.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Metadata
        An entity metadata object.
    """
    return metadata_builder.build(module, **kwargs)


metadata_builder = MetadataBuilder()
metadata_builder.register(ARTF, ArtifactMetadata)
metadata_builder.register(DTIT, DataitemMetadata)
metadata_builder.register(FUNC, FunctionMetadata)
metadata_builder.register(PROJ, ProjectMetadata)
metadata_builder.register(RUNS, RunMetadata)
metadata_builder.register(TASK, TaskMetadata)
metadata_builder.register(WKFL, WorkflowMetadata)



"""
Dataitem module.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities._base.entity import Entity
from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.stores.builder import get_default_store, get_store
from digitalhub_core.utils.api import api_ctx_create, api_ctx_update
from digitalhub_core.utils.commons import DTIT
from digitalhub_core.utils.exceptions import EntityError
from digitalhub_core.utils.file_utils import clean_all, get_dir
from digitalhub_core.utils.generic_utils import build_uuid, get_timestamp
from digitalhub_core.utils.io_utils import write_yaml
from digitalhub_core.utils.uri_utils import get_extension, map_uri_scheme

if typing.TYPE_CHECKING:
    import pandas as pd
    from digitalhub_core.context.context import Context
    from digitalhub_core.entities.dataitems.metadata import DataitemMetadata
    from digitalhub_core.entities.dataitems.spec import DataitemSpec
    from digitalhub_core.entities.dataitems.status import DataitemStatus


class Dataitem(Entity):
    """
    A class representing a dataitem.
    """

    def __init__(
        self,
        uuid: str,
        kind: str,
        metadata: DataitemMetadata,
        spec: DataitemSpec,
        status: DataitemStatus,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        uuid : str
            UUID.
        kind : str
            Kind of the object.
        metadata : DataitemMetadata
            Metadata of the object.
        spec : DataitemSpec
            Specification of the object.
        status : DataitemStatus
            Status of the object.
        """
        super().__init__()
        self.id = uuid
        self.kind = kind
        self.metadata = metadata
        self.spec = spec
        self.status = status

        self.project = self.metadata.project
        self.name = self.metadata.name
        self.embedded = self.metadata.embedded
        self._obj_attr.extend(["project", "name", "embedded"])

    #############################
    #  Save / Export
    #############################

    def save(self, update: bool = False) -> dict:
        """
        Save dataitem into backend.

        Parameters
        ----------
        uuid : str
            Specify uuid for the dataitem to update

        Returns
        -------
        dict
            Mapping representation of Dataitem from backend.
        """
        obj = self.to_dict()

        if not update:
            api = api_ctx_create(self.metadata.project, DTIT)
            return self._context().create_object(obj, api)

        self.metadata.updated = obj["metadata"]["updated"] = get_timestamp()
        api = api_ctx_update(self.metadata.project, DTIT, self.metadata.name, self.id)
        return self._context().update_object(obj, api)

    def export(self, filename: str | None = None) -> None:
        """
        Export object as a YAML file.

        Parameters
        ----------
        filename : str
            Name of the export YAML file. If not specified, the default value is used.

        Returns
        -------
        None
        """
        obj = self.to_dict()
        filename = filename if filename is not None else f"dataitem_{self.metadata.project}_{self.metadata.name}.yaml"
        write_yaml(filename, obj)

    #############################
    #  Dataitem Methods
    #############################

    def as_df(self, file_format: str | None = None, **kwargs) -> pd.DataFrame:
        """
        Read dataitem as a pandas DataFrame. If the dataitem is not local, it will be downloaded
        to a temporary folder and deleted after the method is executed. If no file_format is passed,
        the function will try to infer it from the dataitem.spec.path attribute.
        The path of the dataitem is specified in the spec attribute, and must be a store aware path.
        If the dataitem is stored on s3 bucket, the path must be s3://<bucket>/<path_to_dataitem>.
        If the dataitem is stored on database (Postgres is the only one supported), the path must
        be sql://postgres/<database>/<schema>/<table/view>.

        Parameters
        ----------
        file_format : str
            Format of the file. (Supported csv and parquet).
        **kwargs
            Keyword arguments.

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame.
        """
        if self.spec.path is None:
            raise EntityError("Path is not specified.")

        store = get_store(self.spec.path)
        tmp_path = False

        # Check file format
        extension = self._get_extension(self.spec.path, file_format)

        # Download dataitem if not local
        if not self._check_local(self.spec.path):
            path = store.download(self.spec.path)
            tmp_path = True
        else:
            path = self.spec.path

        df = store.read_df(path, extension, **kwargs)
        if tmp_path:
            clean_all(get_dir(path))

        return df

    def write_df(self, target_path: str | None = None, df: pd.DataFrame | None = None, **kwargs) -> str:
        """
        Write pandas DataFrame as parquet.
        If no target_path is passed, the dataitem will be written into the default store.
        If no DataFrame is passed, the dataitem will be written into the target_path.

        Parameters
        ----------
        target_path : str
            Path to write the dataframe to
        df : pd.DataFrame
            DataFrame to write.
        **kwargs
            Keyword arguments.

        Returns
        -------
        str
            Path to the written dataframe.
        """
        if target_path is None:
            store = get_default_store()
        else:
            store = get_store(target_path)
        if df is None:
            df = self.as_df()
        return store.write_df(df, target_path, **kwargs)

    #############################
    #  Context
    #############################

    def _context(self) -> Context:
        """
        Get context.

        Returns
        -------
        Context
            Context.
        """
        return get_context(self.metadata.project)

    #############################
    #  Helper Methods
    #############################

    @staticmethod
    def _check_local(path: str) -> bool:
        """
        Check if path is local.

        Parameters
        ----------
        path : str
            Path to check.

        Returns
        -------
        bool
            True if local, False otherwise.
        """
        return map_uri_scheme(path) == "local"

    @staticmethod
    def _get_extension(path: str, file_format: str | None = None) -> str:
        """
        Get extension of path.

        Parameters
        ----------
        path : str
            Path to get extension from.
        file_format : str
            File format.

        Returns
        -------
        str
            File extension.

        Raises
        ------
        EntityError
            If file format is not supported.
        """
        if file_format is not None:
            return file_format
        scheme = map_uri_scheme(path)
        if scheme == "sql":
            return "parquet"
        ext = get_extension(path)
        if ext is not None:
            return ext
        raise EntityError("Unknown file format. Only csv and parquet are supported.")


def dataitem_from_parameters(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    embedded: bool = True,
    key: str | None = None,
    path: str | None = None,
    **kwargs,
) -> Dataitem:
    """
    Create a new object instance.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        Identifier of the dataitem.
    kind : str
        The type of the dataitem.
    uuid : str
        UUID.
    description : str
        Description of the dataitem.
    embedded : bool
        Flag to determine if object must be embedded in project.
    key : str
        Representation of the dataitem, e.g. store://etc.
    path : str
        Path to the dataitem on local file system or remote storage.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Dataitem
       Object instance.
    """
    uuid = build_uuid(uuid)
    key = key if key is not None else f"store://{project}/dataitems/{kind}/{name}:{uuid}"
    metadata = build_metadata(
        DTIT,
        project=project,
        name=name,
        version=uuid,
        description=description,
        embedded=embedded,
    )
    spec = build_spec(
        DTIT,
        kind,
        key=key,
        path=path,
        **kwargs,
    )
    status = build_status(DTIT)
    return Dataitem(
        uuid=uuid,
        kind=kind,
        metadata=metadata,
        spec=spec,
        status=status,
    )


def dataitem_from_dict(obj: dict) -> Dataitem:
    """
    Create dataitem from dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create dataitem from.

    Returns
    -------
    Dataitem
        Dataitem object.
    """
    return Dataitem.from_dict(DTIT, obj)



"""
Artifact operations module.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities.artifacts.entity import artifact_from_dict, artifact_from_parameters
from digitalhub_core.utils.api import api_ctx_delete, api_ctx_read, api_ctx_update
from digitalhub_core.utils.commons import ARTF
from digitalhub_core.utils.entities_utils import parse_entity_key
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.artifacts.entity import Artifact


def create_artifact(**kwargs) -> Artifact:
    """
    Create a new artifact with the provided parameters.

    Parameters
    ----------
    **kwargs
        Keyword arguments.

    Returns
    -------
    Artifact
        Object instance.
    """
    return artifact_from_parameters(**kwargs)


def create_artifact_from_dict(obj: dict) -> Artifact:
    """
    Create a new Artifact instance from a dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create the Artifact from.

    Returns
    -------
    Artifact
        Artifact object.
    """
    return artifact_from_dict(obj)


def new_artifact(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    embedded: bool = True,
    key: str | None = None,
    src_path: str | None = None,
    target_path: str | None = None,
    **kwargs,
) -> Artifact:
    """
    Create an instance of the Artifact class with the provided parameters.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        Identifier of the artifact.
    kind : str
        The type of the artifact.
    uuid : str
        UUID.
    description : str
        Description of the artifact.
    embedded : bool
        Flag to determine if object must be embedded in project.
    key : str
        Representation of artfact like store://etc..
    src_path : str
        Path to the artifact on local file system.
    targeth_path : str
        Destination path of the artifact.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Artifact
       Object instance.
    """
    obj = create_artifact(
        project=project,
        name=name,
        kind=kind,
        uuid=uuid,
        description=description,
        embedded=embedded,
        key=key,
        src_path=src_path,
        target_path=target_path,
        **kwargs,
    )
    obj.save()
    return obj


def get_artifact(project: str, name: str, uuid: str | None = None) -> Artifact:
    """
    Retrieves artifact details from the backend.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the artifact.
    uuid : str
        UUID.

    Returns
    -------
    Artifact
        Object instance.
    """
    api = api_ctx_read(project, ARTF, name, uuid=uuid)
    obj = get_context(project).read_object(api)
    return artifact_from_dict(obj)


def get_artifact_from_key(key: str) -> Artifact:
    """
    Get artifact from key.

    Parameters
    ----------
    key : str
        Key of the artifact.
        It's format is store://<project>/artifacts/<kind>/<name>:<uuid>.
    """
    project, name, uuid = parse_entity_key(key)
    return get_artifact(project, name, uuid)


def import_artifact(file: str) -> Artifact:
    """
    Import an Artifact object from a file using the specified file path.

    Parameters
    ----------
    file : str
        Path to the file.

    Returns
    -------
    Artifact
        Object instance.
    """
    obj = read_yaml(file)
    return artifact_from_dict(obj)


def delete_artifact(project: str, name: str, uuid: str | None = None) -> dict:
    """
    Delete artifact from the backend. If the uuid is not specified, delete all versions.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the artifact.
    uuid : str
        UUID.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_ctx_delete(project, ARTF, name, uuid=uuid)
    return get_context(project).delete_object(api)


def update_artifact(artifact: Artifact) -> dict:
    """
    Update a artifact.

    Parameters
    ----------
    artifact : Artifact
        The artifact to update.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_ctx_update(artifact.metadata.project, ARTF, artifact.metadata.name, uuid=artifact.id)
    return get_context(artifact.metadata.project).update_object(artifact.to_dict(), api)



"""
Base Function specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.spec import Spec
from pydantic import BaseModel


class FunctionSpec(Spec):
    """
    Specification for a Function.
    """

    def __init__(
        self,
        source: str | None = None,
        **kwargs,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        source : str
            Path to the Function's source code on the local file system.
        **kwargs
            Keyword arguments.
        """
        self.source = source

        self._any_setter(**kwargs)


class FunctionParams(BaseModel):
    """
    Function parameters model.
    """

    source: str = None
    """Path to the Function's source code on the local file system."""



"""
Workflow base specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.spec import Spec
from pydantic import BaseModel


class WorkflowSpec(Spec):
    """
    Workflow specifications.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """


class WorkflowParams(BaseModel):
    """
    Workflow parameters.
    """



"""
Context module.
"""
from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.projects.entity import Project


class Context:
    """
    The context for a project. It contains the project name and the client.
    It exposes CRUD operations for the entities.
    The context is created by the context builder.
    """

    def __init__(self, project: Project) -> None:
        """
        Constructor.

        Parameters
        ----------
        project : Project
            The project to create the context for.
        """
        self.name = project.metadata.name
        self.client = project._client
        self.local = project._client.is_local()

    def create_object(self, obj: dict, api: str) -> dict:
        """
        Create an object.

        Parameters
        ----------
        obj : dict
            The object to create.
        api : str
            The api to create the object with.

        Returns
        -------
        dict
            The created object.
        """
        return self.client.create_object(obj, api)

    def read_object(self, api: str) -> dict:
        """
        Get an object.

        Parameters
        ----------
        api : str
            The api to get the object with.

        Returns
        -------
        dict
            The read object.
        """
        return self.client.read_object(api)

    def update_object(self, obj: dict, api: str) -> dict:
        """
        Update an object.

        Parameters
        ----------
        obj : dict
            The object to update.
        api : str
            The api to update the object with.

        Returns
        -------
        dict
            The updated object.
        """
        return self.client.update_object(obj, api)

    def delete_object(self, api: str) -> dict:
        """
        Delete an object.

        Parameters
        ----------
        api : str
            The api to delete the object with.

        Returns
        -------
        dict
            The deleted object.
        """
        return self.client.delete_object(api)



"""
FunctionStatus class module.
"""
from __future__ import annotations
from digitalhub_core.entities._base.status import Status


class FunctionStatus(Status):
    """
    Status class for function entities.
    """



"""
Common URI utils.
"""
from __future__ import annotations

from pathlib import Path
from urllib.parse import ParseResult, urlparse


def parse_uri(uri: str) -> ParseResult:
    """
    Parse an uri.

    Parameters
    ----------
    uri : str
        URI.

    Returns
    -------
    ParseResult
        ParseResult object.
    """
    return urlparse(uri)


def get_uri_scheme(uri: str) -> str:
    """
    Get scheme of an URI.

    Parameters
    ----------
    uri : str
        URI.

    Returns
    -------
    str
        URI scheme.
    """
    return parse_uri(uri).scheme


def get_uri_netloc(uri: str) -> str:
    """
    Return URI netloc/bucket.

    Parameters
    ----------
    uri : str
        URI.

    Returns
    -------
    str
        URI netloc.
    """
    return parse_uri(uri).netloc


def get_uri_path(uri: str) -> str:
    """
    Return URI path.

    Parameters
    ----------
    uri : str
        URI.

    Returns
    -------
    str
        URI path.
    """
    return parse_uri(uri).path


def get_name_from_uri(uri: str) -> str:
    """
    Return filename from uri.

    Parameters
    ----------
    uri : str
        URI.

    Returns
    -------
    str
        Filename.
    """
    return Path(uri).name


def get_extension(uri: str) -> str:
    """
    Return extension from uri.

    Parameters
    ----------
    uri : str
        URI.

    Returns
    -------
    str
        Extension.
    """
    return Path(uri).suffix[1:]


def build_key(dst: str) -> str:
    """
    Build key to upload objects.

    Parameters
    ----------
    dst : str
        Destination URI.

    Returns
    -------
    str
        Key.
    """
    key = get_uri_path(dst)
    if key.startswith("/"):
        key = key[1:]
    return key


def as_uri(path: str) -> str:
    """
    Convert a path to an URI.

    Parameters
    ----------
    path : str
        Path.

    Returns
    -------
    str
        URI.

    Notes
    -----
    If path is a relative path, it will be returned as it is.
    """
    try:
        return Path(path).as_uri()
    except Exception as exc:
        if "relative path can't be expressed as a file URI" in exc.args[0]:
            return path
        raise exc


def map_uri_scheme(uri: str) -> str:
    """
    Map an URI scheme to a common scheme.

    Parameters
    ----------
    uri : str
        URI.

    Returns
    -------
    str
        Mapped scheme type.

    Raises
    ------
    ValueError
        If the scheme is unknown.
    """
    scheme = get_uri_scheme(uri)
    if scheme in ["", "file", "local"]:
        return "local"
    if scheme in ["http", "https", "remote"]:
        return "remote"
    if scheme in ["s3", "s3a", "s3n"]:
        return "s3"
    if scheme in ["sql", "postgresql"]:
        return "sql"
    raise ValueError(f"Unknown scheme '{scheme}'!")



"""
Spec factory entity.
"""
from __future__ import annotations

import importlib
import typing

from digitalhub_core.entities.artifacts.spec import ArtifactParams, ArtifactSpec
from digitalhub_core.entities.dataitems.spec import DataitemParams, DataitemSpec
from digitalhub_core.entities.projects.spec import ProjectParams, ProjectSpec
from digitalhub_core.entities.runs.spec import RunParams, RunSpec
from digitalhub_core.entities.workflows.spec import WorkflowParams, WorkflowSpec
from digitalhub_core.utils.exceptions import EntityError
from digitalhub_core.utils.modules_utils import import_registry
from pydantic import ValidationError

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.spec import Spec


REGISTRY = {
    "artifact": [ArtifactSpec, ArtifactParams],
    "dataitem": [DataitemSpec, DataitemParams],
    "project": [ProjectSpec, ProjectParams],
    "run": [RunSpec, RunParams],
    "workflow": [WorkflowSpec, WorkflowParams],
}


def build_spec(
    entity: str, kind: str, ignore_validation: bool = False, module_kind: str | None = None, **kwargs
) -> Spec:
    """
    Build runtimes.

    Parameters
    ----------
    entity : str
        Type of entity.
    kind : str
        The type of Spec to build.
    ignore_validation : bool
        Whether to ignore the validation of the parameters.
    module_kind : str, optional
        The module to import (for function and tasks).
    **kwargs
        Keyword arguments.


    Returns
    -------
    Spec
        A Spec object with the given parameters.
    """
    try:
        if module_kind is not None:
            # Import registry
            registry = import_registry(module_kind)

            # Get name of module and class
            obj = registry.get_spec(entity, kind)

            # Import spec and model
            module = importlib.import_module(obj.module)
            class_spec = getattr(module, obj.class_spec)
            class_params = getattr(module, obj.class_params)
        else:
            # Get spec and model
            class_spec, class_params = REGISTRY[kind]

        # Validate arguments
        if not ignore_validation:
            kwargs = class_params(**kwargs).dict()

        return class_spec(**kwargs)
    except (ModuleNotFoundError, ImportError):
        raise ValueError(f"Runtime '{kind}' not found")
    except KeyError:
        raise EntityError(f"Unsupported parameters kind '{kind}' for entity {entity}")
    except ValidationError as err:
        raise EntityError(f"Invalid parameters for kind '{kind}' for entity {entity}") from err



"""
Artifact module.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities._base.entity import Entity
from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.stores.builder import get_store
from digitalhub_core.utils.api import api_ctx_create, api_ctx_update
from digitalhub_core.utils.commons import ARTF
from digitalhub_core.utils.exceptions import EntityError
from digitalhub_core.utils.file_utils import check_file
from digitalhub_core.utils.generic_utils import build_uuid, get_timestamp
from digitalhub_core.utils.io_utils import write_yaml
from digitalhub_core.utils.uri_utils import get_name_from_uri, map_uri_scheme

if typing.TYPE_CHECKING:
    from digitalhub_core.context.context import Context
    from digitalhub_core.entities.artifacts.metadata import ArtifactMetadata
    from digitalhub_core.entities.artifacts.spec import ArtifactSpec
    from digitalhub_core.entities.artifacts.status import ArtifactStatus


class Artifact(Entity):
    """
    A class representing a artifact.
    """

    def __init__(
        self,
        uuid: str,
        kind: str,
        metadata: ArtifactMetadata,
        spec: ArtifactSpec,
        status: ArtifactStatus,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        uuid : str
            UUID.
        kind : str
            Kind of the object.
        metadata : ArtifactMetadata
            Metadata of the object.
        spec : ArtifactSpec
            Specification of the object.
        status : ArtifactStatus
            Status of the object.
        """
        super().__init__()
        self.id = uuid
        self.kind = kind
        self.metadata = metadata
        self.spec = spec
        self.status = status

        self.project = self.metadata.project
        self.name = self.metadata.name
        self.embedded = self.metadata.embedded
        self._obj_attr.extend(["project", "name", "embedded"])

    #############################
    #  Save / Export
    #############################

    def save(self, update: bool = False) -> dict:
        """
        Save artifact into backend.

        Parameters
        ----------
        update : bool
            Flag to indicate update.

        Returns
        -------
        dict
            Mapping representation of Artifact from backend.
        """
        obj = self.to_dict()

        if not update:
            api = api_ctx_create(self.metadata.project, ARTF)
            return self._context().create_object(obj, api)

        self.metadata.updated = obj["metadata"]["updated"] = get_timestamp()
        api = api_ctx_update(self.metadata.project, ARTF, self.metadata.name, self.id)
        return self._context().update_object(obj, api)

    def export(self, filename: str | None = None) -> None:
        """
        Export object as a YAML file.

        Parameters
        ----------
        filename : str
            Name of the export YAML file. If not specified, the default value is used.

        Returns
        -------
        None
        """
        obj = self.to_dict()
        filename = filename if filename is not None else f"artifact_{self.metadata.project}_{self.metadata.name}.yaml"
        write_yaml(filename, obj)

    #############################
    #  Context
    #############################

    def _context(self) -> Context:
        """
        Get context.

        Returns
        -------
        Context
            Context.
        """
        return get_context(self.metadata.project)

    #############################
    #  Artifacts Methods
    #############################

    def as_file(self, target: str | None = None) -> str:
        """
        Get artifact as file. In the case of a local store, the store returns the current
        path of the artifact. In the case of a remote store, the artifact is downloaded in
        a temporary directory.

        Parameters
        ----------
        target : str
            Target path is the remote path of the artifact where it is stored

        Returns
        -------
        str
            Path of the artifact (temporary or not).
        """
        # Check if target path is provided and if it is remote
        trg = self._parameter_or_default(target, self.spec.target_path)
        self._check_remote(trg)

        # If local store, return local artifact path
        # Check if source path is provided and if it is local
        store = get_store(trg)
        if store.is_local():
            src = self._parameter_or_default(None, self.spec.src_path)
            self._check_local(src)
            return src

        return store.download(trg)

    def download(self, target: str | None = None, dst: str | None = None, overwrite: bool = False) -> str:
        """
        Download artifact from backend.

        Parameters
        ----------
        target : str
            Target path is the remote path of the artifact
        dst : str
            Destination path as filename
        overwrite : bool
            Specify if overwrite an existing file

        Returns
        -------
        str
            Path of the downloaded artifact.
        """

        # Check if target path is provided and if it is remote
        trg = self._parameter_or_default(target, self.spec.target_path)
        self._check_remote(trg)

        # Check if download destination path is specified and rebuild it if necessary
        dst = dst if dst is not None else f"./{get_name_from_uri(trg)}"

        # Check if destination path exists for overwrite
        self._check_overwrite(dst, overwrite)

        # Download artifact and return path
        store = get_store(trg)
        return store.download(trg, dst)

    def upload(self, source: str | None = None, target: str | None = None) -> str:
        """
        Upload artifact to backend.

        Parameters
        ----------
        source : str
            Source path is the local path of the artifact
        target : str
            Target path is the remote path of the artifact

        Returns
        -------
        str
            Path of the uploaded artifact.
        """
        # Check if target path is provided and if it is remote
        trg = self._parameter_or_default(target, self.spec.target_path)
        self._check_remote(trg)

        # Check if source path is provided and if it is local
        src = self._parameter_or_default(source, self.spec.src_path)
        self._check_local(src)

        # Get store
        store = get_store(trg)
        if store.is_local():
            raise EntityError("Cannot target local store for upload.")

        # Upload artifact and return remote path
        return store.upload(src, trg)

    #############################
    #  Private Helpers
    #############################

    @staticmethod
    def _parameter_or_default(parameter: str | None = None, default: str | None = None) -> str:
        """
        Check whether a parameter is specified or not. If not, return the default value. If also
        the default value is not specified, raise an exception. If parameter is specified, but
        default value is not, return the parameter and set the default value to the parameter.

        Parameters
        ----------
        parameter : str
            Parameter to check.
        default : str
            Default value.

        Returns
        -------
        str
            Parameter or default value.

        Raises
        ------
        EntityError
            If parameter and default value are not specified.
        """
        if parameter is None:
            if default is None:
                raise EntityError("Path is not specified.")
            return default
        return parameter

    @staticmethod
    def _check_local(path: str) -> None:
        """
        Check through URI scheme if given path is local or not.

        Parameters
        ----------
        path : str
            Path of some source.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If source path is not local.
        """
        if map_uri_scheme(path) != "local":
            raise EntityError("Only local paths are supported for source paths.")

    @staticmethod
    def _check_remote(path: str) -> None:
        """
        Check through URI scheme if given path is remote.

        Parameters
        ----------
        path : str
            Path of some source.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If source path is local.
        """
        if map_uri_scheme(path) == "local":
            raise EntityError("Only remote paths are supported for target paths.")

    @staticmethod
    def _check_overwrite(dst: str, overwrite: bool) -> None:
        """
        Check if destination path exists for overwrite.

        Parameters
        ----------
        dst : str
            Destination path as filename.
        overwrite : bool
            Specify if overwrite an existing file.

        Returns
        -------
        None

        Raises
        ------
        Exception
            If destination path exists and overwrite is False.
        """
        if check_file(dst) and not overwrite:
            raise EntityError(f"File {dst} already exists.")


def artifact_from_parameters(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    embedded: bool = True,
    key: str | None = None,
    src_path: str | None = None,
    target_path: str | None = None,
    **kwargs,
) -> Artifact:
    """
    Create an instance of the Artifact class with the provided parameters.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        Identifier of the artifact.
    kind : str
        The type of the artifact.
    uuid : str
        UUID.
    description : str
        Description of the artifact.
    embedded : bool
        Flag to determine if object must be embedded in project.
    key : str
        Representation of artfact like store://etc..
    src_path : str
        Path to the artifact on local file system.
    targeth_path : str
        Destination path of the artifact.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Artifact
        Artifact object.
    """
    uuid = build_uuid(uuid)
    metadata = build_metadata(
        ARTF,
        project=project,
        name=name,
        version=uuid,
        description=description,
        embedded=embedded,
    )
    key = key if key is not None else f"store://{project}/artifacts/{kind}/{name}:{uuid}"
    spec = build_spec(
        ARTF,
        kind,
        key=key,
        src_path=src_path,
        target_path=target_path,
        **kwargs,
    )
    status = build_status(ARTF)
    return Artifact(
        uuid=uuid,
        kind=kind,
        metadata=metadata,
        spec=spec,
        status=status,
    )


def artifact_from_dict(obj: dict) -> Artifact:
    """
    Create artifact from dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create artifact from.

    Returns
    -------
    Artifact
        Artifact object.
    """
    return Artifact.from_dict(ARTF, obj)



"""
DataitemStatus class module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.status import Status


class DataitemStatus(Status):
    """
    Status class for dataitem entities.
    """



"""
Abstract entity module.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod

from digitalhub_core.entities._base.base import ModelObj
from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.utils.generic_utils import build_uuid


class Entity(ModelObj, metaclass=ABCMeta):
    """
    Abstract class for entities.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """
        # Attributes to render as dict
        self._obj_attr = [
            "id",
            "kind",
            "metadata",
            "spec",
            "status",
        ]
        self._essential_attr = ["kind", "metadata"]

    @abstractmethod
    def save(self, update: bool = False) -> dict:
        """
        Abstract save method.
        """

    @abstractmethod
    def export(self, filename: str | None = None) -> None:
        """
        Abstract save method.
        """

    def to_dict(self, include_all_non_private: bool = False) -> dict:
        """
        Return object as dict with all keys.

        Parameters
        ----------
        include_all_non_private : bool
            Whether to include all non-private attributes. If False, only
            attributes in the _obj_attr list will be included.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        dict_ = super().to_dict()
        if include_all_non_private:
            return dict_
        return {k: v for k, v in dict_.items() if k in self._obj_attr}

    def to_dict_essential(self) -> dict:
        """
        Return object as dict with some attributes.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        dict_ = super().to_dict()
        return {k: v for k, v in dict_.items() if k in self._essential_attr}

    @classmethod
    def from_dict(cls, entity: str, obj: dict) -> "Entity":
        """
        Create object instance from a dictionary.

        Parameters
        ----------
        entity : str
            Entity type.
        obj : dict
            Dictionary to create object from.

        Returns
        -------
        Workflow
            Self instance.
        """
        parsed_dict = cls._parse_dict(entity, obj)
        return cls(**parsed_dict)

    @staticmethod
    def _parse_dict(entity: str, obj: dict) -> dict:
        """
        Get dictionary and parse it to a valid entity dictionary.

        Parameters
        ----------
        entity : str
            Entity type.
        obj : dict
            Dictionary to parse.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        uuid = build_uuid(obj.get("id"))
        kind = obj.get("kind")
        metadata = build_metadata(entity, **obj.get("metadata"))
        spec = build_spec(entity, kind, ignore_validation=True, **obj.get("spec"))
        status = build_status(entity, **obj.get("status"))
        return {
            "uuid": uuid,
            "kind": kind,
            "metadata": metadata,
            "spec": spec,
            "status": status,
        }

    def __repr__(self) -> str:
        """
        Return string representation of the entity object.

        Returns
        -------
        str
            A string representing the entity instance.
        """
        return str(self.to_dict(include_all_non_private=True))



"""
General utilities module.
"""
from __future__ import annotations

import base64
import os
from datetime import datetime
from uuid import uuid4

from digitalhub_core.utils.io_utils import read_text


def build_uuid(uuid: str | None = None) -> str:
    """
    Create a uuid if not given

    Parameters
    ----------
    uuid : str
        UUID. Optional.

    Returns
    -------
    str
        The uuid.
    """
    if uuid is not None:
        return uuid
    return str(uuid4())


def get_timestamp() -> str:
    """
    Get the current timestamp timezoned.

    Returns
    -------
    str
        The current timestamp.
    """
    return datetime.now().astimezone().isoformat()


def decode_string(string: str) -> str:
    """
    Decode a string from base64.

    Parameters
    ----------
    string : str
        The string to decode.

    Returns
    -------
    str
        The string decoded from base64.
    """
    return base64.b64decode(string).decode()


def encode_string(string: str) -> str:
    """
    Encode a string in base64.

    Parameters
    ----------
    string : str
        The string to encode.

    Returns
    -------
    str
        The string encoded in base64.
    """
    return base64.b64encode(string.encode()).decode()


def encode_source(path: str) -> str:
    """
    Read a file and encode in base64 the content.

    Parameters
    ----------
    path : str
        The file path to read.

    Returns
    -------
    str
        The file content encoded in base64.
    """
    return encode_string(read_text(path))


def set_dhub_env(
    endpoint: str | None = None,
    user: str | None = None,
    password: str | None = None,
    token: str | None = None,
) -> None:
    """
    Function to set environment variables for DHub Core config.

    Parameters
    ----------
    endpoint : str
        The endpoint of DHub Core.
    user : str
        The user of DHub Core.
    password : str
        The password of DHub Core.
    token : str
        The token of DHub Core.

    Returns
    -------
    None
    """
    os.environ["DHUB_CORE_ENDPOINT"] = endpoint
    os.environ["DHUB_CORE_USER"] = user
    os.environ["DHUB_CORE_PASSWORD"] = password
    os.environ["DHUB_CORE_TOKEN"] = token



"""
Base Runtime module.
"""
from __future__ import annotations

from abc import abstractmethod


class Runtime:
    """
    Base Runtime class.
    """

    @abstractmethod
    def build(self, function: dict, task: dict, run: dict) -> dict:
        """
        Build run spec.
        """

    @abstractmethod
    def run(self, run: dict) -> dict:
        """
        Execute run task.
        """

    @staticmethod
    def _get_action(run: dict) -> str:
        """
        Get action from run.
        """
        return run.get("spec").get("task").split(":")[0].split("+")[1]



"""
General utilities module.
"""
from __future__ import annotations

import re


def parse_entity_key(key: str) -> tuple[str, str, str]:
    """
    Parse the entity key.

    Parameters
    ----------
    key : str
        The entity key.

    Returns
    -------
    tuple[str, str, str]
        The project, the name and the uuid of the entity.
    """
    pattern: str = r"store://(?P<project>\w+)/(.*)/(?P<name>\w+):(?P<uuid>[A-Za-z0-9\-\_]+)"
    match = re.match(pattern, key)
    if match is None:
        raise ValueError("Invalid key format.")
    project = match.group("project")
    name = match.group("name")
    uuid = match.group("uuid")
    return project, name, uuid



"""
Entity specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.base import ModelObj
from pydantic import BaseModel


class Spec(ModelObj):
    """
    A class representing the specification of an entity.
    """

    @classmethod
    def from_dict(cls, obj: dict) -> "Spec":
        """
        Return entity specification object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity specification.

        Returns
        -------
        EntitySpec
            An entity specification object.
        """
        return cls(**obj)


class SpecRegistry(dict):
    """
    Specification registry generic class.
    """

    def register(self, kind: str, spec: Spec, model: BaseModel) -> None:
        """
        Register for object specification.

        Parameters
        ----------
        kind : Enum
            Object kind.
        spec : Spec
            Object specification.
        model : BaseModel
            Object model.

        Returns
        -------
        None
        """
        self[kind] = {"spec": spec, "model": model}



"""
S3Store module.
"""
from __future__ import annotations

import typing
from io import BytesIO
from typing import Type

import boto3
import botocore.client  # pylint: disable=unused-import
from botocore.exceptions import ClientError
from digitalhub_core.stores.objects.base import Store, StoreConfig
from digitalhub_core.utils.exceptions import StoreError
from digitalhub_core.utils.uri_utils import get_uri_path

if typing.TYPE_CHECKING:
    import pandas as pd


# Type aliases
S3Client = Type["botocore.client.S3"]


class S3StoreConfig(StoreConfig):
    """
    S3 store configuration class.
    """

    endpoint_url: str
    """S3 endpoint URL."""

    aws_access_key_id: str
    """AWS access key ID."""

    aws_secret_access_key: str
    """AWS secret access key."""

    bucket_name: str
    """S3 bucket name."""


class S3Store(Store):
    """
    S3 store class. It implements the Store interface and provides methods to fetch and persist
    artifacts on S3 based storage.
    """

    def __init__(self, name: str, store_type: str, config: S3StoreConfig) -> None:
        """
        Constructor.

        Parameters
        ----------
        config : S3StoreConfig
            S3 store configuration.

        See Also
        --------
        Store.__init__
        """
        super().__init__(name, store_type)
        self.config = config

    ############################
    # IO methods
    ############################

    def download(self, src: str, dst: str | None = None) -> str:
        """
        Download an artifact from S3 based storage.

        See Also
        --------
        fetch_artifact
        """
        return self._registry.get(src, self.fetch_artifact(src, dst))

    def fetch_artifact(self, src: str, dst: str | None = None) -> str:
        """
        Fetch an artifact from S3 based storage. If the destination is not provided,
        a temporary directory will be created and the artifact will be saved there.

        Parameters
        ----------
        src : str
            The source location of the artifact.
        dst : str
            The destination of the artifact on local filesystem.

        Returns
        -------
        str
            Returns the path of the downloaded artifact.
        """
        dst = dst if dst is not None else self._build_temp(src)
        key = self._get_key(src)
        return self._download_file(key, dst)

    def upload(self, src: str, dst: str | None = None) -> str:
        """
        Upload an artifact to S3 based storage.

        See Also
        --------
        persist_artifact
        """
        return self.persist_artifact(src, dst)

    def persist_artifact(self, src: str, dst: str | None = None) -> str:
        """
        Persist an artifact on S3 based storage. If the destination is not provided,
        the key will be extracted from the source path.

        Parameters
        ----------
        src : Any
            The source object to be persisted. It can be a file path as a string or Path object.

        dst : str
            The destination partition for the artifact.

        Returns
        -------
        str
            Returns the URI of the artifact on S3 based storage.
        """
        key = self._get_key(dst) if dst is not None else self._get_key(src)
        return self._upload_file(src, key)

    def write_df(self, df: pd.DataFrame, dst: str | None = None, **kwargs) -> str:
        """
        Write a dataframe to S3 based storage. Kwargs are passed to df.to_parquet().

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe.
        dst : str
            The destination path on S3 based storage.
        **kwargs
            Keyword arguments.

        Returns
        -------
        str
            The path S3 path where the dataframe was saved.
        """
        if dst is None or not dst.endswith(".parquet"):
            dst = f"s3://{self._get_bucket()}/artifacts/data.parquet"
        fileobj = BytesIO()
        df.to_parquet(fileobj, index=False, **kwargs)
        key = self._get_key(dst)
        return self._upload_fileobj(fileobj, key)

    ############################
    # Private helper methods
    ############################

    def _get_bucket(self) -> str:
        """
        Get the name of the S3 bucket from the URI.

        Returns
        -------
        str
            The name of the S3 bucket.
        """
        return str(self.config.bucket_name)

    def _get_client(self) -> S3Client:
        """
        Get an S3 client object.

        Returns
        -------
        S3Client
            Returns a client object that interacts with the S3 storage service.
        """
        cfg = {
            "endpoint_url": self.config.endpoint_url,
            "aws_access_key_id": self.config.aws_access_key_id,
            "aws_secret_access_key": self.config.aws_secret_access_key,
        }
        return boto3.client("s3", **cfg)

    @staticmethod
    def _get_key(path: str) -> str:
        """
        Build key.

        Parameters
        ----------
        path : str
            The source path to get the key from.

        Returns
        -------
        str
            The key.
        """
        key = get_uri_path(path)
        if key.startswith("/"):
            key = key[1:]
        return key

    def _check_factory(self) -> tuple[S3Client, str]:
        """
        Check if the S3 bucket is accessible by sending a head_bucket request.

        Returns
        -------
        tuple[S3Client, str]
            A tuple containing the S3 client object and the name of the S3 bucket.
        """
        client = self._get_client()
        bucket = self._get_bucket()
        self._check_access_to_storage(client, bucket)
        return client, bucket

    def _check_access_to_storage(self, client: S3Client, bucket: str) -> None:
        """
        Check if the S3 bucket is accessible by sending a head_bucket request.

        Parameters
        ----------
        client : S3Client
            The S3 client object.
        bucket : str
            The name of the S3 bucket.

        Returns
        -------
        None

        Raises
        ------
        StoreError:
            If access to the specified bucket is not available.
        """
        try:
            client.head_bucket(Bucket=bucket)
        except ClientError as exc:
            raise StoreError("No access to s3 bucket!") from exc

    def _download_file(self, key: str, dst: str) -> str:
        """
        Download a file from S3 based storage. The function checks if the bucket is accessible
        and if the destination directory exists. If the destination directory does not exist,
        it will be created.

        Parameters
        ----------
        key : str
            The key of the file on S3 based storage.
        dst : str
            The destination of the file on local filesystem.

        Returns
        -------
        str
            The path of the downloaded file.
        """
        client, bucket = self._check_factory()
        self._check_local_dst(dst)
        client.download_file(bucket, key, dst)
        return dst

    def _upload_file(self, src: str, key: str) -> str:
        """
        Upload a file to S3 based storage. The function checks if the bucket is accessible.

        Parameters
        ----------
        src : str
            The source path of the file on local filesystem.
        key : str
            The key of the file on S3 based storage.

        Returns
        -------
        str
            The URI of the uploaded file on S3 based storage.
        """
        client, bucket = self._check_factory()
        client.upload_file(Filename=src, Bucket=bucket, Key=key)
        return f"s3://{bucket}/{key}"

    def _upload_fileobj(self, fileobj: BytesIO, key: str) -> str:
        """
        Upload a fileobject to S3 based storage. The function checks if the bucket is accessible.

        Parameters
        ----------
        fileobj : BytesIO
            The fileobject to be uploaded.
        key : str
            The key of the file on S3 based storage.

        Returns
        -------
        str
            The URI of the uploaded fileobject on S3 based storage.
        """
        client, bucket = self._check_factory()
        client.put_object(Bucket=bucket, Key=key, Body=fileobj.getvalue())
        return f"s3://{bucket}/{key}"

    ############################
    # Store interface methods
    ############################

    @staticmethod
    def is_local() -> bool:
        """
        Check if the store is local.

        Returns
        -------
        bool
            False
        """
        return False



"""
Context builder module.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.context import Context

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.projects.entity import Project


class ContextBuilder:
    """
    The context builder. It implements the builder pattern to create a context instance.
    It allows to use multiple projects at the same time.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """
        self._instances: dict[str, Context] = {}

    def build(self, project_object: Project) -> None:
        """
        Add a project to the context.

        Parameters
        ----------
        project : Project
            The project to add.

        Returns
        -------
        None
        """
        if project_object.metadata.name not in self._instances:
            self._instances[project_object.metadata.name] = Context(project_object)

    def get(self, project: str) -> Context:
        """
        Get a project from the context.

        Parameters
        ----------
        project : str
            The project name.

        Returns
        -------
        Context
            The project context.
        """
        ctx = self._instances.get(project)
        if ctx is None:
            raise ValueError(f"Context '{project}' not found.")
        return ctx

    def remove(self, project: str) -> None:
        """
        Remove a project from the context.

        Parameters
        ----------
        project : str
            The project name.

        Returns
        -------
        None
        """
        self._instances.pop(project, None)


def set_context(project: Project) -> None:
    """
    Wrapper for ContextBuilder.build().

    Parameters
    ----------
    project : Project
        The project object used to set the current context.

    Returns
    -------
    None
    """
    context_builder.build(project)


def get_context(project: str) -> Context:
    """
    Wrapper for ContextBuilder.get().

    Parameters
    ----------
    project : str
        Name of the project.

    Returns
    -------
    Context
        The context for the given project name.
    """
    return context_builder.get(project)


def delete_context(project: str) -> None:
    """
    Wrapper for ContextBuilder.remove().

    Parameters
    ----------
    project : str
        Name of the project.

    Returns
    -------
    None
    """
    context_builder.remove(project)


context_builder = ContextBuilder()



"""
Project module.
"""
from __future__ import annotations

import typing
from typing import Callable, TypeVar

from digitalhub_core.client.builder import get_client
from digitalhub_core.context.builder import set_context
from digitalhub_core.entities._base.entity import Entity
from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.entities.artifacts.crud import (
    create_artifact_from_dict,
    delete_artifact,
    get_artifact,
    new_artifact,
)
from digitalhub_core.entities.dataitems.crud import (
    create_dataitem_from_dict,
    delete_dataitem,
    get_dataitem,
    new_dataitem,
)
from digitalhub_core.entities.functions.crud import (
    create_function_from_dict,
    delete_function,
    get_function,
    new_function,
)
from digitalhub_core.entities.workflows.crud import (
    create_workflow_from_dict,
    delete_workflow,
    get_workflow,
    new_workflow,
)
from digitalhub_core.utils.api import api_base_create, api_base_update
from digitalhub_core.utils.commons import ARTF, DTIT, FUNC, PROJ, WKFL
from digitalhub_core.utils.exceptions import BackendError, EntityError
from digitalhub_core.utils.generic_utils import build_uuid, get_timestamp
from digitalhub_core.utils.io_utils import write_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.artifacts.entity import Artifact
    from digitalhub_core.entities.dataitems.entity import Dataitem
    from digitalhub_core.entities.functions.entity import Function
    from digitalhub_core.entities.projects.metadata import ProjectMetadata
    from digitalhub_core.entities.projects.spec import ProjectSpec
    from digitalhub_core.entities.projects.status import ProjectStatus
    from digitalhub_core.entities.workflows.entity import Workflow

    Entities = TypeVar("Entities", Artifact, Function, Workflow, Dataitem)

LIST = [ARTF, FUNC, WKFL, DTIT]


def constructor_from_dict(
    dto: str,
) -> Callable[[dict], Artifact | Function | Workflow | Dataitem]:
    """
    Get constructor for dto.

    Parameters
    ----------
    dto : str
        DTO.

    Returns
    -------
    Callable[[dict], Artifact | Function | Workflow | Dataitem]
        Constructor from dict.

    Raises
    ------
    EntityError
        If dto is not valid.
    """
    if dto == ARTF:
        return create_artifact_from_dict
    if dto == FUNC:
        return create_function_from_dict
    if dto == WKFL:
        return create_workflow_from_dict
    if dto == DTIT:
        return create_dataitem_from_dict
    raise EntityError(f"DTO {dto} is not valid.")


class Project(Entity):
    """
    A class representing a project.
    """

    def __init__(
        self,
        uuid: str,
        kind: str,
        metadata: ProjectMetadata,
        spec: ProjectSpec,
        status: ProjectStatus,
        local: bool = False,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        uuid : str
            UUID.
        kind : str
            Kind of the object.
        metadata : ProjectMetadata
            Metadata of the object.
        spec : ProjectSpec
            Specification of the object.
        status : ProjectStatus
            Status of the object.
        local: bool
            If True, export locally.
        """
        super().__init__()
        self.id = uuid
        self.kind = kind
        self.metadata = metadata
        self.spec = spec
        self.status = status

        # Private attributes
        self._client = get_client(local)

        self.name = self.metadata.name
        self._obj_attr.append("name")

        # Set context
        set_context(self)

    #############################
    #  Save / Export
    #############################

    def save(self, update: bool = False) -> dict:
        """
        Save project and context into backend.

        Parameters
        ----------
        uuid : bool
            Ignored, placed for compatibility with other objects.

        Returns
        -------
        list
            Mapping representation of Project from backend.
        """
        # Try to refresh project if local client
        if self._client.is_local():
            obj = self._refresh().to_dict()
        else:
            obj = self.to_dict()

        if not update:
            api = api_base_create(PROJ)
            response = self._client.create_object(obj, api)

        self.metadata.updated = obj["metadata"]["updated"] = get_timestamp()
        api = api_base_update(PROJ, self.id)
        response = self._client.update_object(obj, api)
        return response

    def export(self, filename: str | None = None) -> None:
        """
        Export object as a YAML file. If the objects are not embedded, the objects are
        exported as a YAML file.

        Parameters
        ----------
        filename : str
            Name of the export YAML file. If not specified, the default value is used.

        Returns
        -------
        None
        """
        # Try to refresh project if local client
        if self._client.is_local():
            obj = self._refresh().to_dict()
        else:
            obj = self.to_dict()

        obj = self._parse_spec(obj)

        filename = filename if filename is not None else "project.yaml"
        write_yaml(filename, obj)

        # Export objects related to project if not embedded
        for i in LIST:
            for j in self._get_objects(i):
                _obj = constructor_from_dict(i)(j)
                if not _obj.embedded:
                    _obj.export()

    @staticmethod
    def _parse_spec(obj: dict) -> dict:
        """
        Parse spec dictionary.

        Parameters
        ----------
        obj : dict
            Project dictionary.

        Returns
        -------
        dict
            New project dictionary.
        """
        spec = obj.get("spec", {})
        new_spec: dict[str, list] = {}
        for i in spec:
            new_spec[i] = []
            for j in spec[i]:
                if not j.get("embedded", True):
                    _dict = {k: v for k, v in j.items() if k in ["kind", "name", "id"]}
                    new_spec[i].append(_dict)
                else:
                    new_spec[i].append(j)
        obj["spec"] = new_spec
        return obj

    def _refresh(self) -> "Project":
        """
        Refresh object from backend.

        Returns
        -------
        Project
            Project object.
        """
        try:
            api = api_base_update(PROJ, self.metadata.name)
            obj = self._client.read_object(api)
            return self.from_dict(PROJ, obj)
        except BackendError:
            return self

    #############################
    #  Generic operations for objects (artifacts, functions, workflows, dataitems)
    #############################

    def _add_object(self, obj: Entities, kind: str) -> None:
        """
        Add object to project as specification.

        Parameters
        ----------
        obj : Entity
            Object to be added to project.
        kind : str
            Kind of object to be added to project.

        Returns
        -------
        None
        """
        self._check_kind(kind)
        attr = getattr(self.spec, kind, []) + [obj.to_dict()]
        setattr(self.spec, kind, attr)

    def _delete_object(self, name: str, kind: str, uuid: str | None = None) -> None:
        """
        Delete object from project.

        Parameters
        ----------
        name : str
            Name of object to be deleted.
        kind : str
            Kind of object to be deleted.
        uuid : str
            UUID.

        Returns
        -------
        None
        """
        if uuid is None:
            attr_name = "name"
            var = name
        else:
            attr_name = "id"
            var = uuid
        self._check_kind(kind)
        spec_list = getattr(self.spec, kind, [])
        setattr(self.spec, kind, [i for i in spec_list if i.get(attr_name) != var])

    def _get_objects(self, kind: str) -> list[dict]:
        """
        Get dtos objects related to project.

        Parameters
        ----------
        kind : str
            Kind of object to be retrieved.

        Returns
        -------
        list[dict]
            List of objects related to project.

        Raises
        ------
        EntityError
            If kind is not valid.
        """
        self._check_kind(kind)
        return getattr(self.spec, kind, [])

    @staticmethod
    def _check_kind(kind: str) -> None:
        """
        Check if kind is valid.

        Parameters
        ----------
        kind : str
            Kind of object to be checked.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If kind is not valid.
        """
        if kind not in LIST:
            raise EntityError(f"Kind {kind} is not valid.")

    #############################
    #  Artifacts
    #############################

    def new_artifact(self, **kwargs) -> Artifact:
        """
        Create an instance of the Artifact class with the provided parameters.

        Parameters
        ----------
        **kwargs
            Keyword arguments.

        Returns
        -------
        Artifact
           Object instance.
        """
        kwargs["project"] = self.metadata.name
        kwargs["kind"] = "artifact"
        obj = new_artifact(**kwargs)
        self._add_object(obj, ARTF)
        return obj

    def get_artifact(self, name: str, uuid: str | None = None) -> Artifact:
        """
        Get a Artifact from backend.

        Parameters
        ----------
        name : str
            Identifier of the artifact.
        uuid : str
            Identifier of the artifact version.

        Returns
        -------
        Artifact
            Instance of Artifact class.
        """
        obj = get_artifact(
            project=self.metadata.name,
            name=name,
            uuid=uuid,
        )
        self._add_object(obj, ARTF)
        return obj

    def delete_artifact(self, name: str, uuid: str | None = None) -> None:
        """
        Delete a Artifact from project.

        Parameters
        ----------
        name : str
            Identifier of the artifact.
        uuid : str
            Identifier of the artifact version.

        Returns
        -------
        None
        """
        delete_artifact(self.metadata.name, name, uuid=uuid)
        self._delete_object(name, ARTF, uuid=uuid)

    def set_artifact(self, artifact: Artifact) -> None:
        """
        Set a Artifact.

        Parameters
        ----------
        artifact : Artifact
            Artifact to set.

        Returns
        -------
        None
        """
        self._add_object(artifact, ARTF)

    #############################
    #  Functions
    #############################

    def new_function(self, **kwargs) -> Function:
        """
        Create a Function instance with the given parameters.

        Parameters
        ----------
        **kwargs
            Keyword arguments.

        Returns
        -------
        Function
           Object instance.
        """
        kwargs["project"] = self.metadata.name
        obj = new_function(**kwargs)
        self._add_object(obj, FUNC)
        return obj

    def get_function(self, name: str, uuid: str | None = None) -> Function:
        """
        Get a Function from backend.

        Parameters
        ----------
        name : str
            Identifier of the function.
        uuid : str
            Identifier of the function version.

        Returns
        -------
        Function
            Instance of Function class.
        """
        obj = get_function(
            project=self.metadata.name,
            name=name,
            uuid=uuid,
        )
        self._add_object(obj, FUNC)
        return obj

    def delete_function(self, name: str, uuid: str | None = None) -> None:
        """
        Delete a Function from project.

        Parameters
        ----------
        name : str
            Identifier of the function.
        uuid : str
            Identifier of the function version.

        Returns
        -------
        None
        """
        delete_function(self.metadata.name, name, uuid=uuid)
        self._delete_object(name, FUNC, uuid=uuid)

    def set_function(self, function: Function) -> None:
        """
        Set a Function.

        Parameters
        ----------
        function : Function
            Function to set.

        Returns
        -------
        None
        """
        self._add_object(function, FUNC)

    #############################
    #  Workflows
    #############################

    def new_workflow(self, **kwargs) -> Workflow:
        """
        Create a new Workflow instance with the specified parameters.

        Parameters
        ----------
        **kwargs
            Keyword arguments.

        Returns
        -------
        Workflow
            An instance of the created workflow.
        """
        kwargs["project"] = self.metadata.name
        obj = new_workflow(**kwargs)
        self._add_object(obj, WKFL)
        return obj

    def get_workflow(self, name: str, uuid: str | None = None) -> Workflow:
        """
        Get a Workflow from backend.

        Parameters
        ----------
        name : str
            Identifier of the workflow.
        uuid : str
            Identifier of the workflow version.

        Returns
        -------
        Workflow
            Instance of Workflow class.
        """
        obj = get_workflow(
            project=self.metadata.name,
            name=name,
            uuid=uuid,
        )
        self._add_object(obj, WKFL)
        return obj

    def delete_workflow(self, name: str, uuid: str | None = None) -> None:
        """
        Delete a Workflow from project.

        Parameters
        ----------
        name : str
            Identifier of the workflow.
        uuid : str
            Identifier of the workflow version.

        Returns
        -------
        None
        """
        delete_workflow(self.metadata.name, name, uuid=uuid)
        self._delete_object(name, WKFL, uuid=uuid)

    def set_workflow(self, workflow: Workflow) -> None:
        """
        Set a Workflow.

        Parameters
        ----------
        workflow : Workflow
            Workflow to set.

        Returns
        -------
        None
        """
        self._add_object(workflow, WKFL)

    #############################
    #  Dataitems
    #############################

    def new_dataitem(self, **kwargs) -> Dataitem:
        """
        Create a Dataitem.

        Parameters
        ----------
        **kwargs
            Keyword arguments.

        Returns
        -------
        Dataitem
           Object instance.
        """
        kwargs["project"] = self.metadata.name
        kwargs["kind"] = "dataitem"
        obj = new_dataitem(**kwargs)
        self._add_object(obj, DTIT)
        return obj

    def get_dataitem(self, name: str, uuid: str | None = None) -> Dataitem:
        """
        Get a Dataitem from backend.

        Parameters
        ----------
        name : str
            Identifier of the dataitem.
        uuid : str
            Identifier of the dataitem version.

        Returns
        -------
        Dataitem
            Instance of Dataitem class.
        """
        obj = get_dataitem(
            project=self.metadata.name,
            name=name,
            uuid=uuid,
        )
        self._add_object(obj, DTIT)
        return obj

    def delete_dataitem(self, name: str, uuid: str | None = None) -> None:
        """
        Delete a Dataitem from project.

        Parameters
        ----------
        name : str
            Identifier of the dataitem.
        uuid : str
            Identifier of the dataitem version.

        Returns
        -------
        None
        """
        delete_dataitem(self.metadata.name, name, uuid=uuid)
        self._delete_object(name, DTIT, uuid=uuid)

    def set_dataitem(self, dataitem: Dataitem) -> None:
        """
        Set a Dataitem.

        Parameters
        ----------
        dataitem : Dataitem
            Dataitem to set.

        Returns
        -------
        None
        """
        self._add_object(dataitem, DTIT)


def project_from_parameters(
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    local: bool = False,
    context: str = "",
    source: str = "",
    **kwargs,
) -> Project:
    """
    Create project.

    Parameters
    ----------
    name : str
        Identifier of the project.
    kind : str
        The type of the project.
    uuid : str
        UUID.
    description : str
        Description of the project.
    local : bool
        Flag to determine if object will be exported to backend.
    context : str
        The context of the project.
    source : str
        The source of the project.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Project
        Project object.
    """
    uuid = build_uuid(uuid)
    spec = build_spec(
        PROJ,
        kind,
        context=context,
        source=source,
        **kwargs,
    )
    metadata = build_metadata(
        PROJ,
        project=name,
        name=name,
        version=uuid,
        description=description,
    )
    status = build_status(PROJ)
    return Project(
        uuid=uuid,
        kind=kind,
        metadata=metadata,
        spec=spec,
        status=status,
        local=local,
    )


def project_from_dict(obj: dict) -> Project:
    """
    Create project from dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create project from.

    Returns
    -------
    Project
        Project object.
    """
    return Project.from_dict(PROJ, obj)



"""
Module for performing operations on functions.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities.functions.entity import function_from_dict, function_from_parameters
from digitalhub_core.utils.api import api_ctx_delete, api_ctx_read, api_ctx_update
from digitalhub_core.utils.commons import FUNC
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.functions.entity import Function


def create_function(**kwargs) -> Function:
    """
    Create a new object instance.

    Parameters
    ----------
    **kwargs
        Keyword arguments.

    Returns
    -------
    Function
       Object instance.
    """
    return function_from_parameters(**kwargs)


def create_function_from_dict(obj: dict) -> Function:
    """
    Create a new Function instance from a dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create the Function from.

    Returns
    -------
    Function
        Function object.
    """
    return function_from_dict(obj)


def new_function(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    embedded: bool = True,
    source: str | None = None,
    **kwargs,
) -> Function:
    """
    Create a new Function instance and persist it to the backend.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        Identifier of the function.
    kind : str
        The type of the function.
    uuid : str
        UUID.
    description : str
        Description of the function.
    embedded : bool
        Flag to determine if object must be embedded in project.
    source : str
        Path to the function's source code on the local file system.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Function
       Object instance.
    """
    obj = create_function(
        project=project,
        name=name,
        kind=kind,
        uuid=uuid,
        description=description,
        source=source,
        embedded=embedded,
        **kwargs,
    )
    obj.save()
    return obj


def get_function(project: str, name: str, uuid: str | None = None) -> Function:
    """
    Get object from backend.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the function.
    uuid : str
        UUID.

    Returns
    -------
    Function
        Object instance.
    """
    api = api_ctx_read(project, FUNC, name, uuid=uuid)
    obj = get_context(project).read_object(api)
    return function_from_dict(obj)


def import_function(file: str) -> Function:
    """
    Get object from file.

    Parameters
    ----------
    file : str
        Path to the file.

    Returns
    -------
    Function
        Object instance.
    """
    obj = read_yaml(file)
    return function_from_dict(obj)


def delete_function(project: str, name: str, uuid: str | None = None) -> dict:
    """
    Delete function from the backend. If the uuid is not specified, delete all versions.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the function.
    uuid : str
        UUID.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_ctx_delete(project, FUNC, name, uuid=uuid)
    return get_context(project).delete_object(api)


def update_function(function: Function) -> dict:
    """
    Update a function.

    Parameters
    ----------
    function : Function
        The function to update.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_ctx_update(function.metadata.project, FUNC, function.metadata.name, uuid=function.id)
    return get_context(function.metadata.project).update_object(function.to_dict(), api)
